from datetime import datetime

from pydantic import BaseModel, Field, HttpUrl

from maplestory.models.character.cashitem_equip import CashitemOption

from ..types import PetSkill


class PetAutoSkill(BaseModel):
    """펫 버프 자동 스킬 정보

    Attributes:
        skill_1 (str | None): 첫 번째 슬롯에 등록된 자동 스킬
        skill_1_icon (HttpUrl | None): 첫 번째 슬롯에 등록된 자동 스킬 아이콘
        skill_2 (str | None): 두 번째 슬롯에 등록된 자동 스킬
        skill_2_icon (HttpUrl | None): 두 번째 슬롯에 등록된 자동 스킬 아이콘
    """

    skill_1: str | None
    skill_1_icon: HttpUrl | None
    skill_2: str | None
    skill_2_icon: HttpUrl | None


class PetItemOption(CashitemOption):
    """펫 장비 아이템 옵션

    Attributes:
        type: 옵션 타입
        value: 옵션 값
    """


class PetItem(BaseModel):
    """캐릭터 펫 장착 정보

    Attributes:
        name (str | None): 아이템 명
        icon (HttpUrl | None): 아이템 아이콘
        description (str | None): 아이템 설명
        option (list[PetItemOption]): 아이템 표기상 옵션
        scroll_upgrade (int): 업그레이드 횟수
        scroll_upgradable (int): 업그레이드 가능 횟수
        shape (str | None): 아이템 외형
        shape_icon (HttpUrl | None): 아이템 외형 아이콘
    """

    name: str | None = Field(alias="item_name")
    icon: HttpUrl | None = Field(alias="item_icon")
    description: str | None = Field(alias="item_description")
    option: list[PetItemOption] = Field(alias="item_option")
    scroll_upgrade: int
    scroll_upgradable: int
    shape: str | None = Field(alias="item_shape")
    shape_icon: HttpUrl | None = Field(alias="item_shape_icon")


class CharacterPet(BaseModel):
    """캐릭터 펫 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)

        pet_1_name (str | None): 펫1 명
        pet_1_nickname (str | None): 펫1 닉네임
        pet_1_icon (HttpUrl | None): 펫1 아이콘
        pet_1_description (str | None): 펫1 설명
        pet_1_equipment (PetItem | None): 펫1 장착 정보
        pet_1_auto_skill (PetAutoSkill | None): 펫1 펫 버프 자동스킬 정보
        pet_1_pet_type (str | None): 펫1 원더 펫 종류
        pet_1_skill (list[PetSkill]): 펫1 펫 보유 스킬
        pet_1_date_expire (datetime | None): 펫1 마법의 시간 (KST, 시간 단위 데이터로 분은 일괄 0으로 표기)
        pet_1_appearance (str | None): 펫1 외형
        pet_1_appearance_icon (HttpUrl | None): 펫1 외형 아이콘

        pet_2_name (str | None): 펫2 명
        pet_2_nickname (str | None): 펫2 닉네임
        pet_2_icon (HttpUrl | None): 펫2 아이콘
        pet_2_description (str | None): 펫2 설명
        pet_2_equipment (PetItem | None): 펫2 장착 정보
        pet_2_auto_skill (PetAutoSkill | None): 펫2 펫 버프 자동스킬 정보
        pet_2_pet_type (str | None): 펫2 원더 펫 종류
        pet_2_skill (list[PetSkill]): 펫2 펫 보유 스킬
        pet_2_date_expire (datetime | None): 펫2 마법의 시간 (KST, 시간 단위 데이터로 분은 일괄 0으로 표기)
        pet_2_appearance (str | None): 펫2 외형
        pet_2_appearance_icon (HttpUrl | None): 펫2 외형 아이콘

        pet_3_name (str | None): 펫3 명
        pet_3_nickname (str | None): 펫3 닉네임
        pet_3_icon (HttpUrl | None): 펫3 아이콘
        pet_3_description (str | None): 펫3 설명
        pet_3_equipment (PetItem | None): 펫3 장착 정보
        pet_3_auto_skill (PetAutoSkill | None): 펫3 펫 버프 자동스킬 정보
        pet_3_pet_type (str | None): 펫3 원더 펫 종류
        pet_3_skill (list[PetSkill]): 펫3 펫 보유 스킬
        pet_3_date_expire (datetime | None): 펫3 마법의 시간 (KST, 시간 단위 데이터로 분은 일괄 0으로 표기)
        pet_3_appearance (str | None): 펫3 외형
        pet_3_appearance_icon (HttpUrl | None): 펫3 외형 아이콘
    """

    date: datetime = Field(repr=False)

    pet_1_name: str | None
    pet_1_nickname: str | None
    pet_1_icon: HttpUrl | None
    pet_1_description: str | None
    pet_1_equipment: PetItem | None
    pet_1_auto_skill: PetAutoSkill | None
    pet_1_pet_type: str | None
    pet_1_skill: list[PetSkill]
    pet_1_date_expire: datetime | None
    pet_1_appearance: str | None
    pet_1_appearance_icon: HttpUrl | None

    pet_2_name: str | None
    pet_2_nickname: str | None
    pet_2_icon: HttpUrl | None
    pet_2_description: str | None
    pet_2_equipment: PetItem | None
    pet_2_auto_skill: PetAutoSkill | None
    pet_2_pet_type: str | None
    pet_2_skill: list[PetSkill]
    pet_2_date_expire: datetime | None
    pet_2_appearance: str | None
    pet_2_appearance_icon: HttpUrl | None

    pet_3_name: str | None
    pet_3_nickname: str | None
    pet_3_icon: HttpUrl | None
    pet_3_description: str | None
    pet_3_equipment: PetItem | None
    pet_3_auto_skill: PetAutoSkill | None
    pet_3_pet_type: str | None
    pet_3_skill: list[PetSkill]
    pet_3_date_expire: datetime | None
    pet_3_appearance: str | None
    pet_3_appearance_icon: HttpUrl | None



"""이 모듈은 한국 표준시(KST)에 대한 날짜와 시간을 처리하는 함수와 클래스를 제공합니다.

- `AwareDatetime` 클래스는 한국 표준시(KST) 시간대 정보를 요구하는 datetime을 나타냅니다.
- `validate` 함수는 주어진 datetime이 한국 표준시(KST) 시간대 정보를 가지고 있는지 확인합니다.
- `now`, `yesterday`, `datetime` 함수는 각각 현재, 어제, 주어진 시간의 한국 표준시(KST)을 반환합니다.
"""

from datetime import datetime as _datetime
from datetime import timedelta
from typing import Annotated, Any
from zoneinfo import ZoneInfo

from annotated_types import Timezone
from pydantic import GetCoreSchemaHandler
from pydantic.types import _check_annotated_type
from pydantic_core import core_schema

KST_TZ = ZoneInfo("Asia/Seoul")
KST_TZ_CONSTRAINT = 9 * 60 * 60
KSTdatetime = Annotated[_datetime, Timezone("Asia/Seoul")]


class AwareDatetime:
    """한국 표준시(KST) 시간대 정보를 요구하는 datetime을 나타냅니다."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        if cls is source:
            # used directly as a type
            return core_schema.datetime_schema(tz_constraint=KST_TZ_CONSTRAINT)

        schema = handler(source)
        _check_annotated_type(schema["type"], "datetime", cls.__name__)
        schema["tz_constraint"] = KST_TZ_CONSTRAINT
        return schema

    def __repr__(self) -> str:
        return "KstAwareDatetime"


def validate(date: _datetime):
    """주어진 날짜가 한국 표준시(KST) 시간대 정보를 가지고 있는지 확인합니다."""
    if date.tzinfo is None:
        raise ValueError("datetime should have timezone info.")

    if date.tzinfo.utcoffset(date) != timedelta(seconds=KST_TZ_CONSTRAINT):
        raise ValueError("datetime should have KST timezone info.")


def now() -> AwareDatetime:
    """현재 한국 표준시(KST) 시간을 반환합니다."""
    return _datetime.now(KST_TZ)


def yesterday() -> AwareDatetime:
    """어제의 한국 표준시(KST) 시간을 반환합니다."""
    today_kst = _datetime.now(KST_TZ)
    return today_kst - timedelta(days=1)


def datetime(
    year, month=None, day=None, hour=0, minute=0, second=0, microsecond=0
) -> AwareDatetime:
    """주어진 시간의 한국 표준시(KST)를 반환합니다."""
    return _datetime(year, month, day, hour, minute, second, microsecond, tzinfo=KST_TZ)



from datetime import datetime

from pydantic import BaseModel, Field


class CharacterDojang(BaseModel):
    """무릉도장 최고기록 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class (str): 캐릭터 직업
        world_name (str): 월드 명
        dojang_best_floor (int): 무릉도장 최고 기록 층수
        date_dojang_record (datetime | None): 무릉도장 최고 기록 달성 일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        dojang_best_time (int): 무릉도장 최고 층수 클리어에 걸린 시간 (초)

    Notes:
        무릉도장 최고 기록 달성 일자가 오래된 경우 date_dojang_record가 None으로 표기될 수 있습니다.
    """

    date: datetime = Field(repr=False)
    character_class: str
    world_name: str
    dojang_best_floor: int
    date_dojang_record: datetime | None
    dojang_best_time: int



import re
from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field, computed_field, field_validator

from ..types import ArtifactCrystalOption

StatTuple = tuple[str, int | float]
UNION_ARTIFACT_MAX_LEVEL = 10


def parse_stat_string(stat_str: str) -> StatTuple | list[StatTuple]:
    """
    Parse a string containing stat information and return the parsed results.

    Args:
        stat_str (str): The string containing the stat information.

    Returns:
        StatTuple | list[StatTuple]: The parsed results.
            If the stat_str matches the pattern and contains a single stat, a tuple with the stat name and value is returned.
            If the stat_str matches the pattern and contains multiple stats, a list of tuples with the stat names and values is returned.

    Raises:
        ValueError: If the stat_str does not match the pattern.

    Examples:
        >>> parse_stat_string("올스탯 150 증가")
        ('올스탯', 150)

        >>> parse_stat_string("공격력 18, 마력 18 증가")
        [('공격력', 18), ('마력', 18)]

        >>> parse_stat_string("데미지 15.00% 증가")
        ('데미지', 0.15)

        >>> parse_stat_string("보스 몬스터 공격 시 데미지 15.00% 증가")
        ('보스 몬스터 공격 시 데미지', 0.15)

        >>> parse_stat_string("몬스터 방어율 무시 20% 증가")
        ('몬스터 방어율 무시', 0.2)

        >>> parse_stat_string("아이템 드롭률 7% 증가")
        ('아이템 드롭률', 0.07)

        >>> parse_stat_string("크리티컬 데미지 2.40% 증가")
        ('크리티컬 데미지', 0.024)
    """

    stat_pattern = r"([가-힣\s]+)\s(\d+(\.\d+)?%?)?(\s증가)?"

    matches = re.findall(stat_pattern, stat_str)
    if not matches:
        raise ValueError(f"Invalid stat string: {stat_str}")

    results = []
    for match in matches:
        stat_name = match[0].strip()
        stat_value = match[1]

        if stat_value is None:
            raise ValueError(f"Invalid stat string: {stat_str}")
        elif "%" in stat_value:
            stat_value = float(stat_value.rstrip("%")) / 100.0
            results.append((stat_name, stat_value))
        else:
            results.append((stat_name, int(stat_value)))

    return results[0] if len(results) == 1 else results


class UnionArtifactEffect(BaseModel):
    """유니온 아티팩트 효과 정보

    Attributes:
        name: 아티팩트 효과 명
        level: 아티팩트 효과 레벨

    Examples:
        {
            "name": "올스탯 150 증가",
            "level": 10
        },
        {
            "name": "데미지 15.00% 증가",
            "level": 10
        },
        {
            "name": "보스 몬스터 공격 시 데미지 15.00% 증가",
            "level": 10
        },
        {
            "name": "몬스터 방어율 무시 20% 증가",
            "level": 10
        },
        {
            "name": "버프 지속시간 20% 증가",
            "level": 10
        },
        {
            "name": "아이템 드롭률 7% 증가",
            "level": 6
        },
        {
            "name": "크리티컬 확률 20% 증가",
            "level": 10
        },
        {
            "name": "크리티컬 데미지 2.40% 증가",
            "level": 6
        },
        {
            "name": "최대 HP 2250, 최대 MP 2250 증가",
            "level": 3
        },
        {
            "name": "공격력 30, 마력 30 증가",
            "level": 10
        },
        {
            "name": "재사용 대기시간 미적용 확률 7.50% 증가",
            "level": 10
        },
        {
            "name": "추가 경험치 획득 6% 증가, 다수 공격 스킬의 최대 공격 가능 대상 수 1 증가",
            "level": 5
        }
    """

    name: str
    level: int

    @property
    def stat(self) -> StatTuple | list[StatTuple]:
        return parse_stat_string(self.name)

    def is_max_level(self) -> bool:
        return self.level == UNION_ARTIFACT_MAX_LEVEL

    def required_ap_for_max_level(self) -> int:
        # TODO: Implement required AP calculation
        pass


class UnionArtifactCrystal(BaseModel):
    """유니온 아티팩트 크리스탈 정보

    Attributes:
        name: 아티팩트 크리스탈 명
        validity_flag: 능력치 유효 여부 (0:유효, 1:유효하지 않음)
        date_expire: 능력치 유효 기간(KST)
        level: 아티팩트 크리스탈 등급
        option_name_1: 아티팩트 크리스탈 첫 번째 옵션 명
        option_name_2: 아티팩트 크리스탈 두 번째 옵션 명
        option_name_3: 아티팩트 크리스탈 세 번째 옵션 명
    """

    name: str
    validity_flag: str = Field(repr=False)
    date_expire: datetime
    level: int
    option1: ArtifactCrystalOption = Field(alias="crystal_option_name_1", repr=False)
    option2: ArtifactCrystalOption = Field(alias="crystal_option_name_2", repr=False)
    option3: ArtifactCrystalOption = Field(alias="crystal_option_name_3", repr=False)

    @field_validator("date_expire", mode="before")
    @classmethod
    def make_date_expire(cls, v: Any) -> datetime:
        try:
            # 2024/02/20 19:26:00:000
            date = datetime.strptime(v, "%Y-%m-%dT%H:%M%z")
        except ValueError:
            # 2024-01-26T00:00+09:00
            try:
                date = datetime.strptime(v, "%Y/%m/%d %H:%M:%S:%f")
            except ValueError as error:
                raise error

        return date

    @computed_field
    def expired(self) -> bool:
        return self.validity_flag == "1"

    @computed_field(repr=False)
    def active(self) -> bool:
        return self.validity_flag == "0"

    valid = active

    @computed_field
    def options(self) -> list[ArtifactCrystalOption]:
        return [self.option1, self.option2, self.option3]

    @computed_field
    def remaining_time(self) -> int:
        return self.date_expire - datetime.now()

    @computed_field
    def remaining_time_str(self) -> int:
        days = self.remaining_time.days
        seconds = self.remaining_time.seconds
        hours = seconds // 3600
        minutes = (seconds % 3600) // 60
        seconds = seconds % 60
        return f"{days}일 {hours}시간 {minutes}분 {seconds}초"

    @computed_field
    def remaining_time_short_str(self) -> int:
        days = self.remaining_time.days + self.remaining_time.seconds // 3600 / 24
        return f"{days:.1f}일"

    @computed_field
    def remaining_days(self) -> int:
        return (self.date_expire - datetime.now()).days


class UnionArtifactCrystalGrade(Enum):
    GRADE_1 = (0, 0)
    GRADE_2 = (1, 1)
    GRADE_3 = (2, 3)
    GRADE_4 = (2, 5)
    GRADE_5 = (3, 8)

    def __init__(self, consumed_ap, cumulative_ap):
        self.consumed_ap = consumed_ap
        self.cumulative_ap = cumulative_ap

    @property
    def required_ap_for_next_level(self) -> int:
        return self.consumed_ap

    @property
    def required_ap_for_max_level(self) -> int:
        return self.__class__.GRADE_5.cumulative_ap - self.cumulative_ap


class UnionArtifact(BaseModel):
    """유니온 아티팩트 정보

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        effects: 아티팩트 효과 정보
        crystals: 아티팩트 크리스탈 정보
        remain_ap: 잔여 아티팩트 AP
    """

    date: datetime = Field(repr=False)
    effects: list[UnionArtifactEffect] = Field(alias="union_artifact_effect")
    crystals: list[UnionArtifactCrystal] = Field(alias="union_artifact_crystal")
    remain_ap: int | None = Field(alias="union_artifact_remain_ap")



from typing import Literal

ArtifactCrystalOption = Literal[
    "공격력/마력 증가",
    "데미지 증가",
    "몬스터 방어율 무시 증가",
    "버프 지속시간 증가",
    "보스 몬스터 공격 시 데미지 증가",
    "아이템 드롭률 증가",
    "올스탯 증가",
    "재사용 대기시간 미적용 확률 증가",
    "최대 HP/MP 증가",
    "추가 경험치 획득 증가",
    "크리티컬 데미지 증가",
    "크리티컬 확률 증가",
]

# {
#     "name": "공격력 12, 마력 12 증가",
#     "level": 4
# },
# {
#     "name": "데미지 12.00% 증가",
#     "level": 8
# },
# {
#     "name": "보스 몬스터 공격 시 데미지 12.00% 증가",
#     "level": 8
# },
# {
#     "name": "몬스터 방어율 무시 16% 증가",
#     "level": 8
# },
# {
#     "name": "버프 지속시간 8% 증가",
#     "level": 4
# },
# {
#     "name": "재사용 대기시간 미적용 확률 3.00% 증가",
#     "level": 4
# },
# {
#     "name": "크리티컬 확률 20% 증가",
#     "level": 10
# },
# {
#     "name": "크리티컬 데미지 4.00% 증가",
#     "level": 10
# },
# {
#     "name": "추가 경험치 획득 12% 증가, 다수 공격 스킬의 최대 공격 가능 대상 수 1 증가",
#     "level": 10
# }
UnionWorldName = Literal[
    "스카니아",
    "베라",
    "루나",
    "제니스",
    "크로아",
    "유니온",
    "엘리시움",
    "이노시스",
    "레드",
    "오로라",
    "아케인",
    "노바",
    "리부트",
    "리부트2",
]
UnionRaiderStatName = Literal[
    "공격력/마력",
    "데미지 증가",
    "몬스터 방어율 무시 증가",
    "버프 지속시간 증가",
    "보스 몬스터 공격 시 데미지 증가",
    "아이템 드롭률 증가",
    "올스탯 증가",
    "스킬 재사용 대기시간",
    "최대 HP/MP 증가",
    "추가 경험치 획득 증가",
    "크리티컬 데미지 증가",
    "크리티컬 확률 증가",
    "상태 이상 내성",
    "STR",
    "DEX",
    "INT",
    "LUK",
    "STR, DEX, LUK",
    "최대 HP",
]



from typing import Annotated, Literal

from annotated_types import Ge, Le

AbilityNumber = Annotated[int, Ge(1), Le(3)]
CharacterLevel = Annotated[int, Ge(1), Le(300)]
CharacterClassLevel = Literal["0", "1", "2", "3", "4", "5", "6"]
CharacterGender = Literal["남", "여"]
AndroidGender = CharacterGender | None
ItemGender = CharacterGender | None
ItemOptionValue = Annotated[int, Ge(-100)]  # 혼돈의 주문서로 음수 가능
ItemSlotName = Literal[
    "귀고리",
    "기계 심장",
    "눈장식",
    "망토",
    "모자",
    "무기",
    "반지1",
    "반지2",
    "반지3",
    "반지4",
    "뱃지",
    "벨트",
    "보조무기",
    "상의",
    "신발",
    "어깨장식",
    "얼굴장식",
    "엠블렘",
    "장갑",
    "펜던트",
    "펜던트2",
    "포켓 아이템",
    "하의",
    "훈장",
]
LabelName = Literal["스페셜라벨", "레드라벨", "블랙라벨", "마스터라벨"] | None
LinkSkillLevel = Annotated[int, Ge(1), Le(10)]
PetSkill = Literal[
    "버프 스킬 자동 사용",
    "아이템 줍기",
    "이동반경 확대",
    "자동 줍기",
    "펫 자이언트 스킬",
    "펫 훈련 스킬",
    "HP 물약충전",
    "MP 물약충전",
]
HyperStatLevel = Annotated[int, Ge(0), Le(15)]
HyperStatType = Literal[
    "STR",
    "DEX",
    "INT",
    "LUK",
    "HP",
    "MP",
    "DF/TF/PP",
    "크리티컬 확률",
    "크리티컬 데미지",
    "방어율 무시",
    "데미지",
    "보스 몬스터 공격 시 데미지 증가",
    "상태 이상 내성",
    "공격력/마력",
    "획득 경험치",
    "아케인포스",
    "일반 몬스터 공격 시 데미지 증가",
]
VCoreType = Literal["강화코어", "스킬코어", "특수코어"] | None
VCoreLevel = Annotated[int, Ge(1), Le(30)]



from datetime import datetime

from pydantic import BaseModel, Field

from maplestory.utils.repr import HideZeroStatRepresentation

from ..types import CharacterClass


class SymbolItem(HideZeroStatRepresentation, BaseModel):
    """캐릭터 심볼 장비 상세 정보

    Attributes:
        name: 심볼 이름
        icon: 심볼 아이콘
        description: 심볼 설명
        force: 심볼 포스 증가 수치
        level: 심볼 레벨
        str: 심볼로 증가한 힘(STR)
        dex: 심볼로 증가한 민첩(DEX)
        int: 심볼로 증가한 지력(INT)
        luk: 심볼로 증가한 운(LUK)
        hp: 심볼로 증가한 체력(HP)
        growth_count: 현재 보유 성장치
        require_growth_count: 다음 레벨까지 필요 성장치
    """

    name: str = Field(alias="symbol_name")
    icon: str = Field(alias="symbol_icon")
    description: str = Field(alias="symbol_description")
    force: int = Field(alias="symbol_force")
    level: int = Field(alias="symbol_level")
    STR: int = Field(alias="symbol_str")
    DEX: int = Field(alias="symbol_dex")
    INT: int = Field(alias="symbol_int")
    LUK: int = Field(alias="symbol_luk")
    HP: int = Field(alias="symbol_hp")
    growth_count: int = Field(alias="symbol_growth_count")
    require_growth_count: int = Field(alias="symbol_require_growth_count")

    @property
    def cumulative_growth_count(self) -> int:
        # TODO: 누적 성장치
        return 0

    @property
    def available_max_level(self) -> int:
        # TODO: 누적 성장치로 달성 가능한 최대 레벨
        return 0

    @property
    def available_max_level_meso(self) -> int:
        # TODO: 누적 성장치로 달성 가능한 최대 레벨까지 내야 하는 메소 (밀린 세금)
        return 0


class SymbolEquipment(BaseModel):
    """캐릭터 심볼 정보

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class: 캐릭터 직업
        symbol: 심볼 정보
    """

    date: datetime = Field(repr=False)
    character_class: CharacterClass
    symbol: list[SymbolItem]



from __future__ import annotations

from datetime import datetime

import maplestory.utils.kst as kst
from maplestory.enums import QueryableDate

DATE_FORMAT = "%Y-%m-%d"

# BASIC_MINIMUM_DATE = datetime(2023, 12, 21)

# # 캐릭터 정보 조회
# CHARACTER_MINIMUM_DATE = BASIC_MINIMUM_DATE

# # 길드 정보 조회
# GUILD_MINIMUM_DATE = BASIC_MINIMUM_DATE

# # 유니온 정보 조회
# UNION_MINIMUM_DATE = BASIC_MINIMUM_DATE

# # 랭킹 정보 조회
# RANK_MINIMUM_DATE = datetime(2023, 12, 22)

# # 큐브 정보 조회
# CUBE_MINIMUM_DATE = datetime(2022, 11, 25)

# # 스타포스 정보 조회
# STARFORCE_MINIMUM_DATE = datetime(2023, 12, 27)

# # 잠재능력 정보 조회
# POTENTIAL_MINIMUM_DATE = datetime(2024, 1, 25)


def to_string(date: datetime | None, format=DATE_FORMAT) -> str | None:
    return None if date is None else date.strftime(format)


def is_valid(date: kst.AwareDatetime, category: QueryableDate) -> bool:
    kst.validate(date)

    if date < category.value:
        dt = category.value.strftime(DATE_FORMAT)
        raise ValueError(f"{category.name}은 {dt}부터 데이터를 조회할 수 있습니다.")


# def validate(date: Kstkst.AwareDatetime, minimum_date: Kstkst.AwareDatetime) -> bool:
#     """
#     입력된 날짜가 지정된 최소 날짜 이후인지 검증합니다.
#     Validates the date and ensures it is after the minimum allowed date.

#     Args:
#         date (datetime): 검증할 날짜.
#                          The date to be validated.
#         minimum_date (datetime): 최소 날짜.
#                                  The minimum allowed date.

#     Returns:
#         datetime: 검증이 통과된 날짜.
#                   The validated date.

#     Raises:
#         ValueError: 입력된 날짜가 최소 날짜 이전인 경우 발생.
#                     If the date is before the minimum allowed date.

#     Example:
#         >>> validate_date(datetime(2023, 1, 1, tzinfo=ZoneInfo("Asia/Seoul")))
#         ValueError: 2023-12-21부터 데이터를 조회할 수 있습니다.

#         >>> validate_date(datetime(2024, 1, 1, tzinfo=ZoneInfo("Asia/Seoul")))
#         datetime(2024, 1, 1, tzinfo=ZoneInfo("Asia/Seoul"))

#     Description:
#         입력된 날짜가 MINIMUM_DATETIME 이전인 경우 ValueError를 발생시킵니다.
#         그렇지 않은 경우 입력된 날짜를 반환합니다.
#     """

#     if date < minimum_date:
#         raise ValueError(f"{minimum_date.strftime(DATE_FORMAT)}부터 데이터를 조회할 수 있습니다.")

#     return date

# def is_character_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= CHARACTER_MINIMUM_DATE

# def is_guild_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= GUILD_MINIMUM_DATE


# def is_union_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= UNION_MINIMUM_DATE


# def is_rank_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= RANK_MINIMUM_DATE


# def is_cube_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= CUBE_MINIMUM_DATE


# def is_starforce_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= STARFORCE_MINIMUM_DATE


# def is_potential_available_date(date: kst.AwareDatetime) -> bool:
#     kst.validate(date)
#     return date >= POTENTIAL_MINIMUM_DATE



from datetime import datetime

from pydantic import BaseModel, Field


class Popularity(BaseModel):
    """캐릭터 인기도 정보

    Attributes:
        date (datetime): 조회 기준일
        popularity (int): 인기도
    """

    date: datetime = Field(repr=False)
    popularity: int



from datetime import datetime

from pydantic import BaseModel, Field

from maplestory.utils.repr import HideNoneRepresentation


class VMatrixCore(HideNoneRepresentation, BaseModel):
    """V코어 정보

    Attributes:
        slot_id: 슬롯 인덱스
        slot_level: 슬롯 레벨
        v_core_name: 코어 명
        v_core_type: 코어 타입
            "강화코어" | "스킬코어" | "특수코어" | null
        v_core_level: 코어 레벨
        v_core_skill_1: 코어에 해당하는 스킬 명
        v_core_skill_2: (강화 코어인 경우) 코어에 해당하는 두 번째 스킬 명
        v_core_skill_3: (강화 코어인 경우) 코어에 해당하는 세 번째 스킬 명

    Notes:
        슬롯에 장작한 코어가 없는 경우, 코어 명과 코어 레벨도 None으로 표기합니다.
    """

    slot_id: int
    slot_level: int
    v_core_name: str | None
    v_core_type: str | None
    v_core_level: int
    v_core_skill_1: str | None
    v_core_skill_2: str | None
    v_core_skill_3: str | None


class VMatrix(BaseModel):
    """V매트릭스 정보

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class: 캐릭터 직업
        cores: V코어 정보
        remain_slot_upgrade_point: 캐릭터 잔여 매트릭스 강화 포인트
    """

    date: datetime = Field(repr=False)
    character_class: str | None
    cores: list[VMatrixCore] = Field(alias="character_v_core_equipment")
    remain_slot_upgrade_point: int | None = Field(
        alias="character_v_matrix_remain_slot_upgrade_point"
    )



"""
This script checks the Nexon's MapleStory game API for updates and performs necessary updates.
Nexon의 메이플스토리 게임 API를 확인하여 업데이트가 필요한 항목을 찾고, 필요한 업데이트를 수행합니다.

The script operates in the following order:
스크립트는 다음과 같은 순서로 작동합니다:

1. Retrieves the current version information of the API.
   API의 현재 버전 정보를 가져옵니다.
2. Compares it with the API information stored locally to check if an update is needed.
   로컬에 저장된 API 정보와 비교하여 업데이트가 필요한지 확인합니다.
3. If an update is needed, it downloads the new API information and updates the local API information.
   업데이트가 필요한 경우, 새로운 API 정보를 다운로드하고 로컬의 API 정보를 업데이트합니다.
4. Outputs the updated files.
   업데이트된 파일들을 출력합니다.
"""

from __future__ import annotations

import json
import urllib.parse
from datetime import datetime
from pathlib import Path

import httpx
from bs4 import BeautifulSoup
from pydantic import BaseModel, RootModel, field_validator
from rich import print

API_VERSION_URL = "https://openapi.nexon.com/game/maplestory/?id=22"


def decode_url(url):
    """Decodes the given URL."""
    try:
        return urllib.parse.unquote(url)
    except Exception as e:
        print(f"Error decoding URL: {e}")
        return None


def add_api_information_to_file(filepath: Path, api_info: APIInformation):
    """
    Adds 'x-fileName', 'x-updateDate', and 'x-fileUrl' to a YAML file just before the 'servers:' section.

    Args:
        filepath (Path): The path to the YAML file.
        api_info (APIInformation): An instance of APIInformation containing fileName, updateDate, and fileUrl.

    Raises:
        ValueError: If the 'servers:' section is not found in the YAML file.
    """
    with open(filepath, "r", encoding="utf-8") as input_file:
        lines = input_file.readlines()

    try:
        index = lines.index("servers:\n")
    except ValueError as e:
        raise ValueError(
            "The 'servers:' section was not found in the YAML file."
        ) from e

    insert_lines = [
        f"  x-fileName: {api_info.fileName}\n",
        f"  x-updateDate: '{api_info.updateDate}'\n",
        f"  x-fileUrl: '{api_info.fileUrl}'\n",
    ]
    output_lines = [*lines[:index], *insert_lines, *lines[index:]]

    with open(filepath, "w", encoding="utf-8") as f:
        f.writelines(output_lines)


class APIInformation(BaseModel):
    id: int
    gameId: str
    categoryName: str
    filePath: str
    fileName: str
    fileUrl: str
    ordering: int
    createEmpNo: int
    createEmpName: str
    updateEmpNo: int
    updateEmpName: str
    createDate: datetime
    updateDate: datetime
    isVisible: bool

    @field_validator("createEmpName")
    def decode_name(cls, v):
        """Decodes the name."""
        return decode_url(v)

    def download(self, directory: Path) -> Path:
        """Downloads the API information to the given directory."""
        r = httpx.get(self.fileUrl)
        filename = f"{self.id}_{self.categoryName.replace(' ', '_')}.yaml"
        open(directory / filename, "wb").write(r.content)
        add_api_information_to_file(directory / filename, self)
        return directory / filename


class APIResultList(RootModel):
    """Represents a list of API results."""

    root: list[APIInformation]

    def __iter__(self):
        return iter(self.root)

    @property
    def latest_update_date(self):
        """Returns the latest update date among the API results."""
        return max(api.updateDate for api in self.root)

    def download(self, directory: Path) -> list[Path]:
        """Downloads all API results to the given directory."""
        filepaths = []
        for api in self:
            filepath = api.download(directory)
            filepaths.append(filepath)
        return filepaths


def retrieve_api_result() -> dict:
    """
    Fetches and returns the API result from the specified URL.

    Returns:
        dict: Parsed API result.
    """
    try:
        response = httpx.get(API_VERSION_URL)
        response.raise_for_status()  # Raise an exception for HTTP errors
        soup = BeautifulSoup(response.text, "lxml")
        script = soup.select_one("#__NEXT_DATA__")
        script_data = json.loads(script.string)
        return script_data["props"]["pageProps"]["apiResult"]
    except Exception as e:
        raise Exception(f"Failed to fetch API result: {str(e)}")


if __name__ == "__main__":
    PROJECT_DIRECTORY = Path(__file__).parent.parent
    DATA_DIRECTORY = PROJECT_DIRECTORY / "data"
    yaml_files = list(DATA_DIRECTORY.glob("*.yaml"))
    openapi_file = DATA_DIRECTORY / "openapi.json"

    local_api_info = json.load(open(openapi_file, "r"))
    print("Local API Information")
    print(local_api_info)

    api_result = retrieve_api_result()
    print("Web API Information")
    print(api_result)
    web_api_info = APIResultList.model_validate(api_result)

    print(f"Local API Version == Web API Version: {local_api_info == api_result}")
    if local_api_info != api_result:
        print("Update Required")
        # Update openapi.json
        open(openapi_file, "w").write(
            json.dumps(api_result, indent=4, ensure_ascii=False)
        )

        # Download updated yaml files
        print("Downloading updated yaml files...")
        new_yaml_files = web_api_info.download(DATA_DIRECTORY)

        # Print updated files
        print("Updated files:")
        updated_files = [*yaml_files, *new_yaml_files, openapi_file]
        updated_files = [
            str(file.relative_to(PROJECT_DIRECTORY)) for file in sorted(updated_files)
        ]
        print(updated_files)

        open("updated_files.txt", "w").write("\n".join(updated_files))



import base64
from io import BytesIO

import httpx
from PIL import Image


def get_image_from_url(url: str) -> Image.Image:
    """
    지정된 URL에서 이미지를 검색하고 PIL Image 객체로 반환합니다.

    Args:
        url (str): 검색할 이미지의 URL.

    Returns:
        Image.Image: PIL Image 객체로 반환된 검색된 이미지.

    Examples:
        >>> from maplestory.utils.image import get_image
        >>> url = "https://example.com/image.jpg"
        >>> image = get_image(url)
        >>> image.show()
    """
    """
    Get an image from the given URL and return it as a PIL Image object.

    Parameters:
        url (str): The URL of the image to retrieve.

    Returns:
        Image.Image: The retrieved image as a PIL Image object.

    Raises:
        httpx.RequestError: If there is an error while making the HTTP request.
        PIL.UnidentifiedImageError: If the retrieved content is not a valid image.

    Example:
        >>> url = "https://example.com/image.jpg"
        >>> image = get_image(url)
    """
    res = httpx.get(url)
    return Image.open(BytesIO(res.content))


def get_image_from_base64str(base64_string: str) -> Image.Image:
    """
    지정된 base64 문자열에서 이미지를 검색하고 PIL Image 객체로 반환합니다.

    Args:
        base64_string (str): 검색할 이미지의 base64 문자열.

    Returns:
        Image.Image: PIL Image 객체로 반환된 검색된 이미지.

    Examples:
        >>> from maplestory.utils.image import get_image_from_base64str
        >>> base64_string = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAABsElEQVRIS+2VvUoDQRSGv2k0E
    """

    return Image.open(BytesIO(base64.b64decode(base64_string)))



from maplestory.models.character.ability import Ability
from maplestory.models.character.android_equip import AndroidEquipment
from maplestory.models.character.basic import CharacterBasic
from maplestory.models.character.beauty_equip import BeautyEquipment
from maplestory.models.character.cashitem_equip import CashitemEquipment
from maplestory.models.character.character import CharacterId
from maplestory.models.character.dojang import CharacterDojang
from maplestory.models.character.hexamatrix import HexaMatrix
from maplestory.models.character.hexamatrix_stat import HexaMatrixStat
from maplestory.models.character.hyper_stat import HyperStat
from maplestory.models.character.item_equip import CharacterEquipment
from maplestory.models.character.link_skill import CharacterLinkSkill
from maplestory.models.character.pet_equip import CharacterPet
from maplestory.models.character.popularity import Popularity
from maplestory.models.character.propensity import Propensity
from maplestory.models.character.set_effect import SetEffect
from maplestory.models.character.skill import CharacterSkill
from maplestory.models.character.stat import CharacterStat
from maplestory.models.character.symbol_equip import SymbolEquipment
from maplestory.models.character.vmatrix import VMatrix

__all__ = [
    "Ability",
    "AndroidEquipment",
    "CharacterBasic",
    "BeautyEquipment",
    "CashitemEquipment",
    "CharacterId",
    "CharacterDojang",
    "HexaMatrix",
    "HexaMatrixStat",
    "HyperStat",
    "CharacterEquipment",
    "CharacterLinkSkill",
    "CharacterPet",
    "Popularity",
    "Propensity",
    "SetEffect",
    "CharacterSkill",
    "CharacterStat",
    "SymbolEquipment",
    "VMatrix",
]



from datetime import datetime

from pydantic import BaseModel, Field

from maplestory.types.character import CharacterGender

from ..types import WorldName


class CharacterBasic(BaseModel):
    """캐릭터 기본 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기) example: 2023-12-21T00:00+09:00
        character_name (str): 캐릭터 명
        world_name (str): 월드 명
        character_gender (str): 캐릭터 성별
        character_class (str): 캐릭터 직업
        character_class_level (int): 캐릭터 직업 차수
        character_level (int): 캐릭터 레벨
        character_exp (int): 현재 레벨에서 보유한 경험치
        character_exp_rate (str): 현재 레벨에서 경험치 퍼센트
        character_guild_name (str | None): 캐릭터 소속 길드 명
        character_image (str): 캐릭터 외형 이미지
    """

    date: datetime = Field(repr=False)
    character_name: str
    world_name: str
    character_gender: str
    character_class: str
    character_class_level: int
    character_level: int
    character_exp: int
    character_exp_rate: str
    character_guild_name: str | None
    character_image: str

    @property
    def 조회기준일(self) -> str:
        return self.date

    @property
    def 캐릭터명(self) -> str:
        return self.character_name

    @property
    def 월드명(self) -> str:
        return self.world_name

    @property
    def 캐릭터성별(self) -> str:
        return self.character_gender

    @property
    def 캐릭터직업(self) -> str:
        return self.character_class

    @property
    def 캐릭터직업차수(self) -> str:
        return self.character_class_level

    @property
    def 레벨(self) -> int:
        return self.character_level

    @property
    def 현재레벨경험치(self) -> int:
        return self.character_exp

    @property
    def 누적경험치(self) -> int:
        # TODO: 누적 경험치 계산
        return self.character_exp

    @property
    def 현재레벨경험치퍼센트(self) -> str:
        return self.character_exp_rate

    @property
    def 길드(self) -> str:
        return self.character_guild_name

    @property
    def 이미지(self) -> str:
        return self.character_image


# character_image = "https://open.api.nexon.com/static/maplestory/Character/HDLDFHOOPBEBJBCFGPJIADFOMCOHAHLBBNFJGDENBFADHHPMBKKIDKBDNNGDBOJLMFIFFDDHFGOIMDIIAHKHGLDIELPAAHCGCCBKBFGKFOCMONOIKEMAOABKIBEBPHAAKADNLOMJEEHMDDDBJMELIINOAGIBEHMIPIBJKMJMOKAEGAJFNFFEJCGDDAFBBNFLODBGEEMHGNMKIPKHFIFBDCOIBMEDCAALIDLEEHIPCMNBCPDCAJHFKJABPHCCPHII.png"

# https://docs.pydantic.dev/latest/concepts/json_schema/#field-customisation
# num: int = Field(..., gt=0, le=10)
# foo: int = Field(..., gt=0, lt=10)
# character_level: conint(min=1, max=300)
# snap: int = Field(
#     42,
#     title='The Snap',
#     description='this is the value of snap',
#     gt=30,
#     lt=50,
# )


# https://docs.pydantic.dev/1.10/usage/types/#constrained-types



from datetime import datetime

from pydantic import BaseModel, Field


class SetEffectOptionInfo(BaseModel):
    """세트 효과 옵션 정보

    Attributes:
        set_count (int): 세트 효과 레벨 (장비 수)
        set_option (str): 적용 중인 세트 효과
    """

    set_count: int
    set_option: str


class SetEffectInfo(BaseModel):
    """세트 효과 정보

    Attributes:
        set_name (str): 세트 효과 명
        total_set_count (int): 세트 개수 (럭키 아이템 포함)
        set_effect_info (list[SetEffectOptionInfo]): 세트 효과 옵션
    """

    set_name: str
    total_set_count: int
    set_effect_info: list[SetEffectOptionInfo]


class SetEffect(BaseModel):
    """세트 효과 정보

    Attributes:
        date (datetime): 조회 기준일
        set_effect (list[SetEffectInfo]): 세트 효과 정보
    """

    date: datetime = Field(repr=False)
    set_effect: list[SetEffectInfo]



from datetime import datetime

from pydantic import BaseModel, Field


class Propensity(BaseModel):
    """성향 정보

    Attributes:
        date (datetime): 조회 기준일
        charisma_level (int): 카리스마 레벨
        sensibility_level (int): 감성 레벨
        insight_level (int): 통찰력 레벨
        willingness_level (int): 의지 레벨
        handicraft_level (int): 손재주 레벨
        charm_level (int): 매력 레벨
    """

    date: datetime = Field(repr=False)
    charisma_level: int
    sensibility_level: int
    insight_level: int
    willingness_level: int
    handicraft_level: int
    charm_level: int



from datetime import datetime
from typing import Annotated, Literal

from annotated_types import Ge, Le, Len

PageCursor = Annotated[str, Len(64, 64)]

WorldName = Literal[
    "스카니아",
    "베라",
    "루나",
    "제니스",
    "크로아",
    "유니온",
    "엘리시움",
    "이노시스",
    "레드",
    "오로라",
    "아케인",
    "노바",
    "리부트",
    "리부트2",
    "버닝",
    "버닝2",
    "버닝3",
]

CharacterGender = Literal["남", "여"]
ItemGender = CharacterGender | None
AndroidGender = CharacterGender | None
CharacterClass = Literal[
    # 모험가
    "초보자",
    "히어로",
    "팔라딘",
    "다크나이트",
    "아크메이지(불,독)",
    "아크메이지(썬,콜)",
    "비숍",
    "보우마스터",
    "신궁",
    "패스파인더",
    "나이트로드",
    "섀도어",
    "듀얼블레이드",
    "바이퍼",
    "캡틴",
    "캐논슈터",
    # 시그너스 기사단
    "노블레스",
    "소울마스터",
    "플레임위자드",
    "윈드브레이커",
    "나이트워커",
    "스트라이커",
    "미하일",
    # 레지스탕스&데몬
    "시티즌",
    "블래스터",
    "배틀메이지",
    "와일드헌터",
    "메카닉",
    "데몬슬레이어",
    "제논",
    "데몬어벤져",
    # 영웅
    "아란",
    "에반",
    "메르세데스",
    "팬텀",
    "은월",
    "루미너스",
    # 노바
    "카이저",
    "카인",
    "카데나",
    "엔젤릭버스터",
    # 레프
    "아델",
    "일리움",
    "칼리",
    "아크",
    # 아니마
    "라라",
    "호영",
    # 초월자
    "제로",
    # 프렌즈 월드
    "키네시스",
    # 기간 한정
    "핑크빈",
    "예티",
]
# FIXME: RankingCharacterClass 가능한 값들 확인하기
RankingCharacterClass = Literal[
    "궁수",
    "기사단",
    "도적",
    "라라",
    "레지스탕스",
    "루미너스",
    "마법사",
    "메르세데스",
    "아델",
    "아란",
    "아크",
    "에반",
    "엔젤릭버스터",
    "은월",
    "일리움",
    "전사",
    "초월자",
    "카데나",
    "카이저",
    "카인",
    "칼리",
    "팬텀",
    "프렌즈 월드",
    "해적",
    "호영",
]
# FIXME: RankingCharacterSubClass 가능한 값들 확인하기
RankingCharacterSubClass = Literal[
    "",
    "나이트로드",
    "나이트워커",
    "다크나이트",
    "데몬슬레이어",
    "듀얼블레이더",
    "메카닉",
    "미하일",
    "바이퍼",
    "배틀메이지",
    "보우마스터",
    "블래스터",
    "비숍",
    "섀도어",
    "소울마스터",
    "스트라이커",
    "신궁",
    "아크메이지(불,독)",
    "아크메이지(썬,콜)",
    "와일드헌터",
    "윈드브레이커",
    "제논",
    "제로",
    "캐논마스터",
    "캡틴",
    "키네시스",
    "팔라딘",
    "패스파인더",
    "플레임위자드",
    "히어로",
]
JobClass = Literal[
    "초보자-전체 전직",
    "전사-전체 전직",
    "전사-검사",
    "전사-파이터",
    "전사-페이지",
    "전사-스피어맨",
    "전사-크루세이더",
    "전사-나이트",
    "전사-버서커",
    "전사-히어로",
    "전사-팔라딘",
    "전사-다크나이트",
    "마법사-전체 전직",
    "마법사-매지션",
    "마법사-위자드(불,)",
    "마법사-위자드(썬,)",
    "마법사-클레릭",
    "마법사-메이지(불,)",
    "마법사-메이지(썬,)",
    "마법사-프리스트",
    "마법사-아크메이지(불,)",
    "마법사-아크메이지(썬,)",
    "마법사-비숍",
    "궁수-전체 전직",
    "궁수-아처",
    "궁수-헌터",
    "궁수-사수",
    "궁수-레인저",
    "궁수-저격수",
    "궁수-보우마스터",
    "궁수-신궁",
    "궁수-아처(패스파인더)",
    "궁수-에인션트아처",
    "궁수-체이서",
    "궁수-패스파인더",
    "도적-전체 전직",
    "도적-로그",
    "도적-어쌔신",
    "도적-시프",
    "도적-허밋",
    "도적-시프마스터",
    "도적-나이트로드",
    "도적-섀도어",
    "도적-세미듀어러",
    "도적-듀어러",
    "도적-듀얼마스터",
    "도적-슬래셔",
    "도적-듀얼블레이더",
    "해적-전체 전직",
    "해적-해적",
    "해적-인파이터",
    "해적-건슬링거",
    "해적-캐논슈터",
    "해적-버커니어",
    "해적-발키리",
    "해적-캐논블래스터",
    "해적-바이퍼",
    "해적-캡틴",
    "해적-캐논마스터",
    "기사단-전체 전직",
    "기사단-노블레스",
    "기사단-소울마스터",
    "기사단-플레임위자드",
    "기사단-윈드브레이커",
    "기사단-나이트워커",
    "기사단-스트라이커",
    "기사단-미하일",
    "아란-전체 전직",
    "에반-전체 전직",
    "레지스탕스-전체 전직",
    "레지스탕스-시티즌",
    "레지스탕스-배틀메이지",
    "레지스탕스-와일드헌터",
    "레지스탕스-메카닉",
    "레지스탕스-데몬슬레이어",
    "레지스탕스-데몬어벤져",
    "레지스탕스-제논",
    "레지스탕스-블래스터",
    "메르세데스-전체 전직",
    "팬텀-전체 전직",
    "루미너스-전체 전직",
    "카이저-전체 전직",
    "엔젤릭버스터-전체 전직",
    "초월자-전체 전직",
    "초월자-제로",
    "은월-전체 전직",
    "프렌즈 월드-전체 전직",
    "프렌즈 월드-키네시스",
    "카데나-전체 전직",
    "일리움-전체 전직",
    "아크-전체 전직",
    "호영-전체 전직",
    "아델-전체 전직",
    "카인-전체 전직",
    "라라-전체 전직",
    "칼리-전체 전직",
]
ExpireDate = datetime | None
PresetNumber = Annotated[int, Ge(1), Le(3)] | None
Grade = Literal["레전드리", "유니크", "에픽", "레어"]
PotentialGrade = Grade | None
CubePotenialGrade = Literal["레전드리", "유니크", "에픽", "레어", "노멀"]
PotentialOption = str | None
ItemOptionValue = Annotated[int, Ge(-100)]  # 혼돈의 주문서로 음수 가능
ItemSlotName = Literal[
    "귀고리",
    "기계 심장",
    "눈장식",
    "망토",
    "모자",
    "무기",
    "반지1",
    "반지2",
    "반지3",
    "반지4",
    "뱃지",
    "벨트",
    "보조무기",
    "상의",
    "신발",
    "어깨장식",
    "얼굴장식",
    "엠블렘",
    "장갑",
    "펜던트",
    "펜던트2",
    "포켓 아이템",
    "하의",
    "훈장",
]
LabelName = Literal["스페셜라벨", "레드라벨", "블랙라벨", "마스터라벨"] | None
PetSkill = Literal[
    "버프 스킬 자동 사용",
    "아이템 줍기",
    "이동반경 확대",
    "자동 줍기",
    "펫 자이언트 스킬",
    "펫 훈련 스킬",
    "HP 물약충전",
    "MP 물약충전",
]
LinkSkillLevel = Annotated[int, Ge(1), Le(10)]
GuildLevel = Annotated[int, Ge(1), Le(30)]

# CubeOptionName = Literal["최대 MP", "올스탯", "점프력", "STR"]

ArtifactCrystalOption = Literal[
    "공격력/마력 증가",
    "데미지 증가",
    "몬스터 방어율 무시 증가",
    "버프 지속시간 증가",
    "보스 몬스터 공격 시 데미지 증가",
    "아이템 드롭률 증가",
    "올스탯 증가",
    "재사용 대기시간 미적용 확률 증가",
    "최대 HP/MP 증가",
    "추가 경험치 획득 증가",
    "크리티컬 데미지 증가",
    "크리티컬 확률 증가",
]



from .apis.character import *

__all__ = [
    "Character",
    "get_character_id",
    "get_basic_character_info_by_ocid",
    "get_basic_character_info",
    "get_popularity_by_ocid",
    "get_popularity",
    "get_character_stat_by_ocid",
    "get_character_stat",
    "get_character_hyper_stat_by_ocid",
    "get_character_hyper_stat",
    "get_character_propensity_by_ocid",
    "get_character_propensity",
    "get_character_ability_by_ocid",
    "get_character_ability",
    "get_character_equipment_by_ocid",
    "get_character_equipment",
    "get_character_cashitem_equipment_by_ocid",
    "get_character_cashitem_equipment",
    "get_character_symbol_equipment_by_ocid",
    "get_character_symbol_equipment",
    "get_character_set_effect_by_ocid",
    "get_character_set_effect",
    "get_character_beauty_equipment_by_ocid",
    "get_character_beauty_equipment",
    "get_character_android_equipment_by_ocid",
    "get_character_android_equipment",
    "get_character_pet_equipment_by_ocid",
    "get_character_pet_equipment",
    "get_character_skill_by_ocid",
    "get_character_skill",
    "get_character_link_skill_by_ocid",
    "get_character_link_skill",
    "get_character_vmatrix_by_ocid",
    "get_character_vmatrix",
    "get_character_hexamatrix_by_ocid",
    "get_character_hexamatrix",
    "get_character_hexamatrix_stat_by_ocid",
    "get_character_hexamatrix_stat",
    "get_character_dojang_record_by_ocid",
    "get_character_dojang_record",
]



import httpx
from httpx import Request, Response
from tenacity import retry, wait_random_exponential

from maplestory.config import Config
from maplestory.error import APIError, ErrorMessage


# @retry(retry=retry_if_exception_type(IOError))
# @retry(wait=wait_random_exponential(multiplier=1, max=8))
def fetch(url_path: str, query_params: dict | None = None) -> dict:
    """
    Fetches data from the API.

    Args:
        url_path : The API path.
        query_params (dict): The query parameters.

    Returns:
        dict: The response data.

    Raises:
        APIError: If an API error occurs.
    """
    config = Config()

    resp: Response = httpx.get(
        f"{config.url}{url_path}",
        params=query_params,
        headers={
            "x-nxopen-api-key": config.key,
            "Accept": "application/json",
        },
        timeout=config.timeout,
    )

    data = resp.json()
    if error := data.get("error"):
        em = ErrorMessage.model_validate(error)
        raise APIError(em, resp)

    return data



from .account import Account
from .potential import CubeHistory, CubeHistoryInfo, PotentialOption
from .starforce import StarforceHistory, StarforceHistoryEvent, StarforceHistoryInfo

__all__ = [
    "Account",
    "PotentialOption",
    "CubeHistoryInfo",
    "CubeHistory",
    "StarforceHistoryEvent",
    "StarforceHistoryInfo",
    "StarforceHistory",
]



from datetime import datetime
from typing import Any, Optional
from zoneinfo import ZoneInfo

import httpx
from httpx import Request, Response
from pydantic import BaseModel, Field

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.error import ErrorMessage
from maplestory.models.character import (
    Ability,
    AndroidEquipment,
    BeautyEquipment,
    CashitemEquipment,
    CharacterBasic,
    CharacterDojang,
    CharacterEquipment,
    CharacterId,
    CharacterLinkSkill,
    CharacterPet,
    CharacterSkill,
    CharacterStat,
    HexaMatrix,
    HexaMatrixStat,
    HyperStat,
    Popularity,
    Propensity,
    SetEffect,
    SymbolEquipment,
    VMatrix,
)
from maplestory.models.guild import GuildBasic, GuildModel
from maplestory.models.history import CubeHistory, StarforceHistory
from maplestory.models.ranking import (
    AchievementRanking,
    DojangRanking,
    GuildRanking,
    OverallRanking,
    TheSeedRanking,
    UnionRanking,
)
from maplestory.models.union import UnionInfo, UnionRaider
from maplestory.utils.network import fetch


class MapleStoryApi(BaseModel):
    api_key: str
    BASE_URL: str = "https://open.api.nexon.com/"
    timeout: int

    character_name: Optional[str] = Field(None, repr=False)
    character_ocid: Optional[str] = Field(None, repr=False)

    guild_name: Optional[str] = Field(None, repr=False)
    world_name: Optional[str] = Field(None, repr=False)
    oguild_id: Optional[str] = Field(None, repr=False)

    def __init__(self, api_key, timeout: int = 5000):
        super().__init__(
            api_key=api_key,
            timeout=timeout,
        )

    def get_character_id(self, character_name: str) -> CharacterId:
        """
        캐릭터의 식별자(ocid)를 조회합니다.
        Fetches the identifier (ocid) of a character.

        Args:
            character_name : 캐릭터의 이름.
                                  The name of the character.

        Returns:
            Character: 주어진 이름의 캐릭터.
                       The character with the given name.

        Note:
            - 2023년 12월 21일부터 데이터를 조회할 수 있습니다.
            Data can be queried from December 21, 2023.
            - 캐릭터 정보 조회 API는 일자별 데이터로 매일 오전 1시부터 전일 데이터 조회가 가능합니다.
            The character information query API provides daily data, and data for the previous day can be queried from 1 AM every day.
            - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
            The ocid may change due to game content changes. Please be careful when updating services based on ocid.
        """
        path = "/maplestory/v1/id"
        query = {
            "character_name": character_name,
        }
        response = fetch(path, query)

        return CharacterId.model_validate(response)

    def get_basic_character_info(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterBasic:
        """
        캐릭터의 기본 정보를 조회합니다.
        Fetches the basic information of a character.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                    Reference date for the query (KST).

        Returns:
            CharacterBasic: 캐릭터의 기본 정보.
                            The basic information of the character.

        Note:
            - 2023년 12월 21일부터 데이터를 조회할 수 있습니다.
            Data can be queried from December 21, 2023.
            - 캐릭터 정보 조회 API는 일자별 데이터로 매일 오전 1시부터 전일 데이터 조회가 가능합니다. (예를 들어, 12월 22일 데이터를 조회하면 22일 00시부터 23일의 00시 사이의 데이터가 조회됩니다.)
            The character information query API provides daily data, and data for the previous day can be queried from 1 AM every day. (For example, if you query data for December 22, you will get data from 00:00 on the 22nd to 00:00 on the 23rd.)
            - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
            The ocid may change due to game content changes. Please be careful when updating services based on ocid.
        """
        path = "/maplestory/v1/character/basic"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CharacterBasic.model_validate(response)

    def get_character_popularity(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> Popularity:
        """
        캐릭터의 인기도 정보를 조회합니다.
        Fetches the popularity information of a character.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterPopularity: 캐릭터의 인기도 정보.
                                 The popularity information of the character.
        """
        path = "/maplestory/v1/character/popularity"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return Popularity(
            date=api_response.get("date"),
            popularity=api_response.get("popularity"),
        )

    def get_character_stat(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterStat:
        """
        캐릭터의 종합능력치 정보를 조회합니다.
        Fetches the comprehensive ability information of a character.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterStat: 캐릭터의 종합능력치 정보.
                           The comprehensive ability information of the character.
        """
        path = "/maplestory/v1/character/stat"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CharacterStat.model_validate(response)

    def get_character_hyper_stat(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> HyperStat:
        """
        캐릭터의 하이퍼스탯 정보를 조회합니다.
        Fetches the hyper stat information of a character.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterHyperStat: 캐릭터의 하이퍼스탯 정보.
                                The hyper stat information of the character.
        """
        path = "/maplestory/v1/character/hyper-stat"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return HyperStat.model_validate(response)

    def get_character_propensity(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> Propensity:
        """
        캐릭터의 성향 정보를 조회합니다.
        Fetches the propensity information of a character.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterPropensity: 캐릭터의 성향 정보.
                                 The propensity information of the character.
        """
        path = "/maplestory/v1/character/propensity"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return Propensity.model_validate(response)

    def get_character_ability(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> Ability:
        """
        캐릭터의 어빌리티 정보를 조회합니다.
        Fetches the ability information of a character.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterAbility: 캐릭터의 어빌리티 정보.
                              The ability information of the character.
        """
        path = "/maplestory/v1/character/ability"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return Ability.model_validate(response)

    def get_character_equipment(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterEquipment:
        """
        장착한 장비 중 캐시 장비를 제외한 나머지 장비 정보를 조회합니다.
        Fetches the information of equipped items excluding cash items.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterItemEquipment: 캐릭터의 장비 정보.
                                    The equipment information of the character.
        """
        path = "/maplestory/v1/character/item-equipment"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CharacterEquipment.model_validate(response)

    def get_character_cashitem_equipment(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CashitemEquipment:
        """
        장착한 캐시 장비 정보를 조회합니다.
        Fetches the information of equipped cash items.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterCashitemEquipment: 캐릭터의 캐시 장비 정보.
                                        The cash item equipment information of the character.
        """
        path = "/maplestory/v1/character/cashitem-equipment"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CashitemEquipment.model_validate(response)

    def get_character_symbol_equipment(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> SymbolEquipment:
        """
        장착한 심볼 정보를 조회합니다.
        Fetches the equipped symbol information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterSymbolEquipment: 캐릭터의 심볼 장비 정보.
                                      The symbol equipment information of the character.
        """
        path = "/maplestory/v1/character/symbol-equipment"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return SymbolEquipment.model_validate(response)

    def get_character_set_effect(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> SetEffect:
        """
        적용받고 있는 세트 효과 정보를 조회합니다.
        Fetches the applied set effect information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterSetEffect: 캐릭터의 세트 효과 정보.
                                The set effect information of the character.
        """
        path = "/maplestory/v1/character/set-effect"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return SetEffect.model_validate(response)

    def get_character_beauty_equipment(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> BeautyEquipment:
        """
        캐릭터 헤어, 성형, 피부 정보를 조회합니다.
        Fetches the character's hair, plastic surgery, and skin information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterBeautyEquipment: 캐릭터의 미용 장비 정보.
                                      The beauty equipment information of the character.
        """
        path = "/maplestory/v1/character/beauty-equipment"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return BeautyEquipment.model_validate(response)

    def get_character_android_equipment(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> AndroidEquipment:
        """
        장착한 안드로이드 정보를 조회합니다.
        Fetches the equipped android information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterAndroidEquipment: 캐릭터의 안드로이드 장비 정보.
                                       The android equipment information of the character.
        """
        path = "/maplestory/v1/character/android-equipment"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return AndroidEquipment.model_validate(response)

    def get_character_pet_equipment(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterPet:
        """
        장착한 펫 및 펫 스킬, 장비 정보를 조회합니다.
        Fetches the equipped pet and pet skill, equipment information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterPetEquipment: 캐릭터의 펫 장비 정보.
                                   The pet equipment information of the character.
        """
        path = "/maplestory/v1/character/pet-equipment"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CharacterPet.model_validate(response)

    def get_character_skill(
        self,
        character_ocid: str,
        skill_grade: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterSkill:
        """
        캐릭터 스킬과 하이퍼 스킬 정보를 조회합니다.
        Fetches the character's skill and hyper skill information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            skill_grade : 조회하고자 하는 전직 차수.
                               The job advancement grade to query.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterSkill: 캐릭터의 스킬 정보.
                            The skill information of the character.
        """
        path = "/maplestory/v1/character/skill"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
            "character_skill_grade": skill_grade,
        }
        response = fetch(path, query)

        return CharacterSkill.model_validate(response)

    def get_character_link_skill(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterLinkSkill:
        """
        장착 링크 스킬 정보를 조회합니다.
        Fetches the equipped link skill information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterLinkSkill: 캐릭터의 링크 스킬 정보.
                                The link skill information of the character.
        """
        path = "/maplestory/v1/character/link-skill"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CharacterLinkSkill.model_validate(response)

    def get_character_vmatrix(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> VMatrix:
        """
        V매트릭스 슬롯 정보와 장착한 V코어 정보를 조회합니다.
        Fetches the VMatrix slot information and the equipped VCore information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterVMatrix: 캐릭터의 V매트릭스 정보.
                              The VMatrix information of the character.
        """
        path = "/maplestory/v1/character/vmatrix"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return VMatrix.model_validate(response)

    def get_character_hexamatrix(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> HexaMatrix:
        """
        HEXA 매트릭스에 장착한 HEXA 코어 정보를 조회합니다.
        Fetches the HEXA core information equipped in the HEXA matrix.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterHexaMatrix: 캐릭터의 HEXA 매트릭스 정보.
                                 The HEXA matrix information of the character.
        """
        path = "/maplestory/v1/character/hexamatrix"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return HexaMatrix.model_validate(response)

    def get_character_hexamatrix_stat(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> HexaMatrixStat:
        """
        HEXA 매트릭스에 설정한 HEXA 스탯 정보를 조회합니다.
        Fetches the HEXA stat information set in the HEXA matrix.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterHexaMatrixStat: 캐릭터의 HEXA 매트릭스 스탯 정보.
                                     The HEXA matrix stat information of the character.
        """
        path = "/maplestory/v1/character/hexamatrix-stat"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return HexaMatrixStat.model_validate(response)

    def get_character_dojang_record(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> CharacterDojang:
        """
        캐릭터 무릉도장 최고 기록 정보를 조회합니다.
        Fetches the highest record information of the character's Mu Lung Dojang.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            CharacterDojang: 캐릭터의 무릉도장 최고 기록 정보.
                             The highest record information of the character's Mu Lung Dojang.
        """
        path = "/maplestory/v1/character/dojang"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return CharacterDojang.model_validate(response)

    def get_union_info(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> UnionInfo:
        """
        유니온 레벨 및 유니온 등급 정보를 조회합니다.
        Fetches the Union level and Union rank information.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            Union: 유니온의 레벨 및 등급 정보.
                   The level and rank information of the Union.
        """
        path = "/maplestory/v1/user/union"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return UnionInfo.model_validate(response)

    def get_union_raider_info(
        self,
        character_ocid: str,
        date: datetime = kst.yesterday(),
    ) -> UnionRaider:
        """
        유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등 상세 정보를 조회합니다.
        Fetches the detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.

        Args:
            character_ocid : 캐릭터의 식별자(ocid).
                                  The identifier (ocid) of the character.
            date : 조회 기준일 (KST).
                                       Reference date for the query (KST).

        Returns:
            UnionRaider: 유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등의 상세 정보.
                         The detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.
        """
        path = "/maplestory/v1/user/union-raider"
        query = {
            "ocid": character_ocid,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return UnionRaider.model_validate(response)

    def get_guild_id(self, guild_name: str, world_name: str) -> GuildModel:
        """
        The get_guild_id function fetches the guild identifier (gcid) information.

        :param self: Represent the instance of the class
        :param guild_name: str: Specify the name of the guild
        :param world_name: str: Specify the world name
        :return: A guildmodel object
        :doc-author: Trelent
        """
        """
        길드 식별자(gcid) 정보를 조회합니다.
        Fetches the guild identifier (gcid) information.

        Args:
            guild_name : 길드 명. The name of the guild.
            world_name : 월드 명. The name of the world.

        Returns:
            GuildModel: 길드 식별자(gcid) 정보. The guild identifier (gcid) information.
        """
        path = "/maplestory/v1/guild/id"
        query = {
            "guild_name": guild_name,
            "world_name": world_name,
        }
        response = fetch(path, query)

        return GuildModel.model_validate(response)

    def get_guild_basic_info(
        self,
        guild_id: str,
        date: datetime = kst.yesterday(),
    ) -> GuildBasic:
        """
        길드 기본 정보를 조회합니다.
        Fetches the basic information of the guild.

        Args:
            guild_id : 길드 식별자. The identifier of the guild.
            date : 조회 기준일(KST). Reference date for the query (KST).

        Returns:
            GuildBasic: 길드의 기본 정보. The basic information of the guild.
        """
        path = "/maplestory/v1/guild/basic"
        query = {
            "oguild_id": guild_id,
            "date": dates.to_string(date),
        }
        response = fetch(path, query)

        return GuildBasic.model_validate(response)

    def get_cube_usage_history(
        self,
        result_count: int,
        date: datetime | None = None,
        cursor: str | None = None,
    ) -> CubeHistory:
        """
        큐브 사용 결과를 조회합니다.
        Fetches the usage result of the cube.

        Args:
            result_count : 한번에 가져오려는 결과의 갯수. The number of results to fetch at once.
            date : 조회 기준일(KST). Reference date for the query (KST).
            cursor : 페이징 처리를 위한 cursor. Cursor for paging.

        Returns:
            CubeHistory: 큐브 사용 결과. The usage result of the cube.
        """
        path = "/maplestory/v1/history/cube"
        query = {
            "count": result_count,
            "date": dates.to_string(date) if date else date,
            "cursor": cursor,
        }
        response = fetch(path, query)

        return CubeHistory.model_validate(response)

    def get_starforce_history(
        self,
        result_count: int,
        base_date: Optional[datetime] = None,
        page_cursor: Optional[str] = None,
    ) -> StarforceHistory:
        """
        스타포스 강화 결과를 조회합니다.
        Fetches the starforce enhancement results.

        Args:
            result_count : 한번에 가져오려는 결과의 갯수(최소 10, 최대 1000)
                               The number of results to fetch at once (minimum 10, maximum 1000)
            base_date (datetime, optional): 조회 기준일(KST) (cursor가 없는 경우 필수이며 cursor와 함께 사용 불가)
                                            Query base date (KST) (required if cursor is not present and cannot be used with cursor)
            page_cursor (str, optional): 페이징 처리를 위한 cursor (date가 없는 경우 필수이며 date와 함께 사용 불가)
                                         Cursor for pagination (required if date is not present and cannot be used with date)
        """
        path = "/maplestory/v1/history/starforce"
        query = {
            "count": result_count,
            "date": dates.to_string(base_date) if base_date else base_date,
            "cursor": page_cursor,
        }
        response = fetch(path, query)
        return StarforceHistory.model_validate(response)

    def get_overall_ranking(
        self,
        world_name: Optional[str] = None,
        world_type: Optional[int] = None,
        class_name: Optional[str] = None,
        character_id: Optional[str] = None,
        page_number: int = 1,
        base_date: datetime = kst.yesterday(),
    ) -> OverallRanking:
        """
        종합 랭킹 정보를 조회합니다.
        Fetches the overall ranking information.

        Args:
            base_date : 조회 기준일(KST)
                                  Query base date (KST)
            world_name (str, optional): 월드 명
                                        World name
            world_type (int, optional): 월드 타입 (0:일반, 1:리부트) (기본 값은 0이며, world_name 입력 시 미 반영)
                                        World type (0: normal, 1: reboot) (default is 0 and not reflected when world_name is entered)
            class_name (str, optional): 직업 및 전직
                                        Job and pre-job
            character_id (str, optional): 캐릭터 식별자
                                          Character identifier
            page_number : 페이지 번호
                              Page number
        """
        path = "/maplestory/v1/ranking/overall"
        query = {
            "date": dates.to_string(base_date),
            "world_name": world_name,
            "world_type": world_type,
            "class": class_name,
            "ocid": character_id,
            "page": page_number,
        }
        response = fetch(path, query)
        return OverallRanking.model_validate(response)

    def get_union_ranking(
        self,
        world_name: str | None = None,
        ocid: str | None = None,
        page: int = 1,
        date: datetime = kst.yesterday(),
    ) -> UnionRanking:
        """유니온 랭킹 정보를 조회합니다.

        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.

        @param date: 조회 기준일(KST)
        @param world_name: 월드 명
        - 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3

        @param ocid: 캐릭터 식별자
        @param page: 페이지 번호
        """
        path = "/maplestory/v1/ranking/union"
        query = {
            "date": date.to_string(date),
            "world_name": world_name,
            "ocid": ocid,
            "page": page,
        }
        r = fetch(path, query)
        return UnionRanking.model_validate(r)

    def fetch_union_ranking(
        self,
        world: Optional[str] = None,
        character_id: Optional[str] = None,
        page_number: int = 1,
        date: datetime = kst.yesterday(),
    ) -> UnionRanking:
        """
        Fetches the union ranking information.

        Args:
            date : The reference date for the query (KST).
            world (Optional[str]): The name of the world.
                Possible values are: Scania, Bera, Luna, Zenith, Croa, Union, Elysium, Enosis, Red, Aurora, Arcane, Nova, Reboot, Reboot2, Burning, Burning2, Burning3.
            character_id (Optional[str]): The identifier of the character.
            page_number : The page number.

        Returns:
            UnionRanking: The union ranking information.

        Note:
            - Data can be queried from December 22, 2023.
            - Ranking information for the day can be queried from 8:30 AM.
            - Please note that the ocid may change due to game content changes. Please be careful when updating services based on ocid.
        """
        path = "/maplestory/v1/ranking/union"
        query = {
            "date": date.to_string(date),
            "world_name": world,
            "ocid": character_id,
            "page": page_number,
        }
        response = fetch(path, query)
        return UnionRanking.model_validate(response)

    def get_guild_ranking(
        self,
        ranking_type: int,
        world_name: str | None = None,
        guild_name: str | None = None,
        page: int = 1,
        date: datetime = kst.yesterday(),
    ) -> GuildRanking:
        """길드 랭킹 정보를 조회합니다.

        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.

        @param date: 조회 기준일(KST)
        @param ranking_type: 랭킹 타입 (0:주간 명성치, 1:플래그 레이스, 2:지하 수로)

        @param world_name: 월드 명
        - 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3

        @param guild_name: 길드 명
        @param page: 페이지 번호
        """
        path = "/maplestory/v1/ranking/guild"
        query = {
            "date": date.to_string(date),
            "ranking_type": ranking_type,
            "world_name": world_name,
            "guild_name": guild_name,
            "page": page,
        }
        r = fetch(path, query)
        return GuildRanking.model_validate(r)

    def fetch_guild_ranking(
        self,
        ranking_type: int,
        world: Optional[str] = None,
        guild: Optional[str] = None,
        page_number: int = 1,
        date: datetime = kst.yesterday(),
    ) -> GuildRanking:
        """
        Fetches the guild ranking information.

        Args:
            date : The reference date for the query (KST).
            ranking_type : The type of ranking (0: Weekly Fame, 1: Flag Race, 2: Underground Waterway).
            world (Optional[str]): The name of the world.
                Possible values are: Scania, Bera, Luna, Zenith, Croa, Union, Elysium, Enosis, Red, Aurora, Arcane, Nova, Reboot, Reboot2, Burning, Burning2, Burning3.
            guild (Optional[str]): The name of the guild.
            page_number : The page number.

        Returns:
            GuildRanking: The guild ranking information.

        Note:
            - Data can be queried from December 22, 2023.
            - Ranking information for the day can be queried from 8:30 AM.
            - Please note that the ocid may change due to game content changes. Please be careful when updating services based on ocid.
        """
        path = "/maplestory/v1/ranking/guild"
        query = {
            "date": date.to_string(date),
            "ranking_type": ranking_type,
            "world_name": world,
            "guild_name": guild,
            "page": page_number,
        }
        response = fetch(path, query)
        return GuildRanking.model_validate(response)

    def get_dojang_ranking(
        self,
        world_name: str | None = None,
        class_name: str | None = None,
        ocid: str | None = None,
        page: int = 1,
        difficulty: int = 1,
        date: datetime = kst.yesterday(),
    ) -> DojangRanking:
        """무릉도장 랭킹 정보를 조회합니다.

        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.

        @param date: 조회 기준일(KST)
        @param difficulty: 구간 (0:일반, 1:통달)
        @param world_name: 월드 명
        - 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3

        @param class_name: 직업 및 전직
        - 초보자-전체 전직, 전사-전체 전직, 전사-검사, 전사-파이터, 전사-페이지, 전사-스피어맨, 전사-크루세이더, 전사-나이트, 전사-버서커, 전사-히어로, 전사-팔라딘, 전사-다크나이트, 마법사-전체 전직, 마법사-매지션, 마법사-위자드(불,독), 마법사-위자드(썬,콜), 마법사-클레릭, 마법사-메이지(불,독), 마법사-메이지(썬,콜), 마법사-프리스트, 마법사-아크메이지(불,독), 마법사-아크메이지(썬,콜), 마법사-비숍, 궁수-전체 전직, 궁수-아처, 궁수-헌터, 궁수-사수, 궁수-레인저, 궁수-저격수, 궁수-보우마스터, 궁수-신궁, 궁수-아처(패스파인더), 궁수-에인션트아처, 궁수-체이서, 궁수-패스파인더, 도적-전체 전직, 도적-로그, 도적-어쌔신, 도적-시프, 도적-허밋, 도적-시프마스터, 도적-나이트로드, 도적-섀도어, 도적-세미듀어러, 도적-듀어러, 도적-듀얼마스터, 도적-슬래셔, 도적-듀얼블레이더, 해적-전체 전직, 해적-해적, 해적-인파이터, 해적-건슬링거, 해적-캐논슈터, 해적-버커니어, 해적-발키리, 해적-캐논블래스터, 해적-바이퍼, 해적-캡틴, 해적-캐논마스터, 기사단-전체 전직, 기사단-노블레스, 기사단-소울마스터, 기사단-플레임위자드, 기사단-윈드브레이커, 기사단-나이트워커, 기사단-스트라이커, 기사단-미하일, 아란-전체 전직, 에반-전체 전직, 레지스탕스-전체 전직, 레지스탕스-시티즌, 레지스탕스-배틀메이지, 레지스탕스-와일드헌터, 레지스탕스-메카닉, 레지스탕스-데몬슬레이어, 레지스탕스-데몬어벤져, 레지스탕스-제논, 레지스탕스-블래스터, 메르세데스-전체 전직, 팬텀-전체 전직, 루미너스-전체 전직, 카이저-전체 전직, 엔젤릭버스터-전체 전직, 초월자-전체 전직, 초월자-제로, 은월-전체 전직, 프렌즈 월드-전체 전직, 프렌즈 월드-키네시스, 카데나-전체 전직, 일리움-전체 전직, 아크-전체 전직, 호영-전체 전직, 아델-전체 전직, 카인-전체 전직, 라라-전체 전직, 칼리-전체 전직

        @param ocid: 캐릭터 식별자
        @param page: 페이지 번호
        """
        path = "/maplestory/v1/ranking/dojang"
        query = {
            "date": date.to_string(date),
            "world_name": world_name,
            "class": class_name,
            "ocid": ocid,
            "page": page,
            "difficulty": difficulty,
        }
        r = fetch(path, query)
        return DojangRanking.model_validate(r)

    def fetch_dojang_ranking(
        self,
        world: Optional[str] = None,
        job_class: Optional[str] = None,
        character_id: Optional[str] = None,
        page_number: int = 1,
        difficulty_level: int = 1,
        date: datetime = kst.yesterday(),
    ) -> DojangRanking:
        """
        Fetches the Dojang ranking information.

        Args:
            world (Optional[str]): The name of the world.
            job_class (Optional[str]): The name of the job class.
            character_id (Optional[str]): The identifier of the character.
            page_number : The page number.
            difficulty_level : The difficulty level (0: Normal, 1: Mastery).
            date : The reference date for the query (KST).

        Returns:
            DojangRanking: The Dojang ranking information.

        Note:
            - Data can be queried from December 22, 2023.
            - Ranking information can be queried from 8:30 AM.
            - Be careful when updating the service based on ocid, as the ocid may change due to game content changes.
        """
        path = "/maplestory/v1/ranking/dojang"
        query = {
            "date": dates.to_string(date),
            "world_name": world,
            "class": job_class,
            "ocid": character_id,
            "page": page_number,
            "difficulty": difficulty_level,
        }
        response = fetch(path, query)
        return DojangRanking.model_validate(response)

    def get_theseed_ranking(
        self,
        world_name: str | None = None,
        ocid: str | None = None,
        page: int = 1,
        date: datetime = kst.yesterday(),
    ) -> TheSeedRanking:
        """더 시드 랭킹 정보를 조회합니다.

        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.

        @param date: 조회 기준일(KST)
        @param world_name: 월드 명
        - 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3
        @param ocid: 캐릭터 식별자
        @param page: 페이지 번호
        """
        path = "/maplestory/v1/ranking/theseed"
        query = {
            "date": date.to_string(date),
            "world_name": world_name,
            "ocid": ocid,
            "page": page,
        }
        r = fetch(path, query)
        return TheSeedRanking.model_validate(r)

    def fetch_theseed_ranking(
        self,
        world: Optional[str] = None,
        character_id: Optional[str] = None,
        page_number: int = 1,
        date: datetime = kst.yesterday(),
    ) -> TheSeedRanking:
        """
        Fetches the TheSeed ranking information.

        Args:
            world (Optional[str]): The name of the world.
            character_id (Optional[str]): The identifier of the character.
            page_number : The page number.
            date : The reference date for the query (KST).

        Returns:
            TheSeedRanking: The TheSeed ranking information.

        Note:
            - Data can be queried from December 22, 2023.
            - Ranking information can be queried from 8:30 AM.
            - Be careful when updating the service based on ocid, as the ocid may change due to game content changes.
        """
        path = "/maplestory/v1/ranking/theseed"
        query = {
            "date": dates.to_string(date),
            "world_name": world,
            "ocid": character_id,
            "page": page_number,
        }
        response = fetch(path, query)
        return TheSeedRanking.model_validate(response)

    def get_achievement_ranking(
        self, ocid: str | None = None, page: int = 1, date: datetime = kst.yesterday()
    ) -> AchievementRanking:
        """업적 랭킹 정보를 조회합니다.

        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.

        @param date: 조회 기준일(KST)
        @param ocid: 캐릭터 식별자
        @param page: 페이지 번호
        """
        path = "/maplestory/v1/ranking/achievement"
        query = {
            "date": date.to_string(date),
            "ocid": ocid,
            "page": page,
        }
        r = fetch(path, query)
        return AchievementRanking.model_validate(r)

    def get_achievement_rankings(
        self,
        character_id: Optional[str] = None,
        page_number: int = 1,
        date: datetime = kst.yesterday(),
    ) -> AchievementRanking:
        """
        Fetches the achievement ranking information.

        Args:
            character_id (Optional[str]): The identifier of the character.
            page_number : The page number.
            date : The reference date for the query (KST).

        Returns:
            AchievementRanking: The achievement ranking information.

        Note:
            - Data can be queried from December 22, 2023.
            - Ranking information can be queried from 8:30 AM.
            - Be careful when updating the service based on ocid, as the ocid may change due to game content changes.
        """
        path = "/maplestory/v1/ranking/achievement"
        query = {
            "date": dates.to_string(date),
            "ocid": character_id,
            "page": page_number,
        }
        response = fetch_data(path, query)
        return AchievementRanking.model_validate(response)



from pydantic import BaseModel


class GuildModel(BaseModel):
    """길드 식별자(oguild_id) 정보

    Attributes:
        oguild_id: 길드 식별자
    """

    oguild_id: str

    @property
    def id(self) -> str:
        return self.oguild_id



from typing import Annotated, Literal

from annotated_types import Ge, Le

CharacterGender = Literal["남", "여"]
ItemGender = CharacterGender | None
AndroidGender = CharacterGender | None

ItemOptionValue = Annotated[int, Ge(-100)]  # 혼돈의 주문서로 음수 가능
ItemSlotName = Literal[
    "귀고리",
    "기계 심장",
    "눈장식",
    "망토",
    "모자",
    "무기",
    "반지1",
    "반지2",
    "반지3",
    "반지4",
    "뱃지",
    "벨트",
    "보조무기",
    "상의",
    "신발",
    "어깨장식",
    "얼굴장식",
    "엠블렘",
    "장갑",
    "펜던트",
    "펜던트2",
    "포켓 아이템",
    "하의",
    "훈장",
]
LabelName = Literal["스페셜라벨", "레드라벨", "블랙라벨", "마스터라벨"] | None
PetSkill = Literal[
    "버프 스킬 자동 사용",
    "아이템 줍기",
    "이동반경 확대",
    "자동 줍기",
    "펫 자이언트 스킬",
    "펫 훈련 스킬",
    "HP 물약충전",
    "MP 물약충전",
]
LinkSkillLevel = Annotated[int, Ge(1), Le(10)]

CharacterClass = Literal[
    # 모험가
    "초보자",
    "히어로",
    "팔라딘",
    "다크나이트",
    "아크메이지(불,독)",
    "아크메이지(썬,콜)",
    "비숍",
    "보우마스터",
    "신궁",
    "패스파인더",
    "나이트로드",
    "섀도어",
    "듀얼블레이드",
    "바이퍼",
    "캡틴",
    "캐논슈터",
    # 시그너스 기사단
    "노블레스",
    "소울마스터",
    "플레임위자드",
    "윈드브레이커",
    "나이트워커",
    "스트라이커",
    "미하일",
    # 레지스탕스&데몬
    "시티즌",
    "블래스터",
    "배틀메이지",
    "와일드헌터",
    "메카닉",
    "데몬슬레이어",
    "제논",
    "데몬어벤져",
    # 영웅
    "아란",
    "에반",
    "메르세데스",
    "팬텀",
    "은월",
    "루미너스",
    # 노바
    "카이저",
    "카인",
    "카데나",
    "엔젤릭버스터",
    # 레프
    "아델",
    "일리움",
    "칼리",
    "아크",
    # 아니마
    "라라",
    "호영",
    # 초월자
    "제로",
    # 프렌즈 월드
    "키네시스",
    # 기간 한정
    "핑크빈",
    "예티",
]



from __future__ import annotations

from enum import Enum

from httpx import HTTPStatusError, Response
from pydantic import BaseModel, Field, computed_field


# https://stackoverflow.com/questions/50473951/how-can-i-attach-documentation-to-members-of-a-python-enum
class DocEnum(Enum):
    def __new__(cls, value, doc=None):
        self = object.__new__(cls)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self


class APIErrorCode(str, Enum):
    """API 에러 코드를 나타내는 Enum 클래스입니다. 각 코드는 특정 API 에러 상황을 나타냅니다.

    Attributes:
        OPENAPI00001: 서버 내부 오류
        OPENAPI00002: 권한이 없는 경우
        OPENAPI00003: 유효하지 않은 식별자
        OPENAPI00004: 파라미터 누락 또는 유효하지 않음
        OPENAPI00005: 유효하지 않은 API KEY
        OPENAPI00006: 유효하지 않은 게임 또는 API PATH
        OPENAPI00007: API 호출량 초과
        OPENAPI00009: 데이터 준비 중
    """

    """Enum class representing API error codes. 
    
    Each code represents a specific API error situation.
    
    Attributes:
        OPENAPI00001: Server internal error
        OPENAPI00002: No permission
        OPENAPI00003: Invalid identifier
        OPENAPI00004: Missing or invalid parameter
        OPENAPI00005: Invalid API KEY
        OPENAPI00006: Invalid game or API PATH
        OPENAPI00007: API call limit exceeded
        OPENAPI00009: Data preparation in progress
    """
    """API 에러 코드를 나타내는 Enum 클래스입니다. 각 코드는 특정 API 에러 상황을 나타냅니다.
    
    Attributes:
        OPENAPI00001: 서버 내부 오류
        OPENAPI00002: 권한이 없는 경우
        OPENAPI00003: 유효하지 않은 식별자
        OPENAPI00004: 파라미터 누락 또는 유효하지 않음
        OPENAPI00005: 유효하지 않은 API KEY
        OPENAPI00006: 유효하지 않은 게임 또는 API PATH
        OPENAPI00007: API 호출량 초과
        OPENAPI00009: 데이터 준비 중    
    """

    OPENAPI00001 = "서버 내부 오류"
    OPENAPI00002 = "권한이 없는 경우"
    OPENAPI00003 = "유효하지 않은 식별자"
    OPENAPI00004 = "파라미터 누락 또는 유효하지 않음"
    OPENAPI00005 = "유효하지 않은 API KEY"
    OPENAPI00006 = "유효하지 않은 게임 또는 API PATH"
    OPENAPI00007 = "API 호출량 초과"
    OPENAPI00009 = "데이터 준비 중"

    @classmethod
    def from_code(cls, code: str) -> APIErrorCode:
        """
        문자열 코드를 해당 Enum 값으로 변환합니다.

        Args:
            code (str): 변환할 문자열 코드입니다.

        Returns:
            APIErrorCode: 해당하는 Enum 값입니다.
        """
        """
        Converts a string code to its corresponding Enum value.
        
        Args:
            code (str): The string code to convert.
        
        Returns:
            APIErrorCode: The corresponding Enum value.
        """
        return cls._member_map_[code]


class ErrorMessage(BaseModel):
    """에러 메세지

    Attributes:
        code: 에러 코드 (예: OPENAPI00001)
        message: 에러 설명

    Examples:
        >>> error = ErrorMessage(name="OPENAPI00004", message="Please input valid parameter")
        >>> error.code
        'OPENAPI00004'
        >>> error.message
        'Please input valid parameter'
    """

    code: str = Field(alias="name")
    message: str


class ErrorResponse(BaseModel):
    error: ErrorMessage


class APIError(HTTPStatusError):
    """MapleStory API Exception"""

    code: str
    message: str
    status: int

    def __init__(
        self,
        error: ErrorMessage,
        response: Response,
    ):
        super().__init__(
            f"{error.message} ({error.code})",
            request=response.request,
            response=response,
        )
        self.code = error.code
        self.message = error.message
        self.status = response.status_code

    @computed_field
    @property
    def description(self) -> str:
        return APIErrorCode.from_code(self.code).value


# ERROR_LIST = [
#     ("OPENAPI00001", 500, "Internal Server Error", "서버 내부 오류")
#     ("OPENAPI00002", 403, "Forbidden", "권한이 없는 경우")
#     ("OPENAPI00003", 400, "Bad Request", "유효하지 않은 식별자")
#     ("OPENAPI00004", 400, "Bad Request", "파라미터 누락 또는 유효하지 않음")
#     ("OPENAPI00005", 400, "Bad Request", "유효하지 않은 API KEY")
#     ("OPENAPI00006", 400, "Bad Request", "유효하지 않은 게임 또는 API PATH")
#     ("OPENAPI00007", 429, "Too Many Requests", "API 호출량 초과")
#     ("OPENAPI00009", 400, "Bad Request", "데이터 준비 중")
# ]
# class APIException(Exception):
#     def __init__(self, error: ErrorMessage):
#         self.error_code = error.code
#         self.message = error.message
#         super().__init__(error.message)


# ("에러 코드", "응답 코드", "응답명", "설명")
# ("OPENAPI00001", 500, "Internal Server Error", "서버 내부 오류")
# ("OPENAPI00002", 403, "Forbidden", "권한이 없는 경우")
# ("OPENAPI00003", 400, "Bad Request", "유효하지 않은 식별자")
# ("OPENAPI00004", 400, "Bad Request", "파라미터 누락 또는 유효하지 않음")
# ("OPENAPI00005", 400, "Bad Request", "유효하지 않은 API KEY")
# ("OPENAPI00006", 400, "Bad Request", "유효하지 않은 게임 또는 API PATH")
# ("OPENAPI00007", 429, "Too Many Requests", "API 호출량 초과")
# ("OPENAPI00009", 400, "Bad Request", "데이터 준비 중")


# class APIErrorCode(str, Enum):
#     Unauthorized = "unauthorized"
#     """The bearer token is not valid."""

#     RestrictedResource = "restricted_resource"
#     """Given the bearer token used, the client doesn't have permission to
#     perform this operation."""

#     ObjectNotFound = "object_not_found"
#     """Given the bearer token used, the resource does not exist.
#     This error can also indicate that the resource has not been shared with owner
#     of the bearer token."""

#     RateLimited = "rate_limited"
#     """This request exceeds the number of requests allowed. Slow down and try again."""

#     InvalidJSON = "invalid_json"
#     """The request body could not be decoded as JSON."""

#     InvalidRequestURL = "invalid_request_url"
#     """The request URL is not valid."""

#     InvalidRequest = "invalid_request"
#     """This request is not supported."""

#     ValidationError = "validation_error"
#     """The request body does not match the schema for the expected parameters."""

#     ConflictError = "conflict_error"
#     """The transaction could not be completed, potentially due to a data collision.
#     Make sure the parameters are up to date and try again."""

#     InternalServerError = "internal_server_error"
#     """An unexpected error occurred. Reach out to Notion support."""

#     ServiceUnavailable = "service_unavailable"
#     """Notion is unavailable. Try again later.
#     This can occur when the time to respond to a request takes longer than 60 seconds,
#     the maximum request timeout."""


# ("에러 코드", "응답 코드", "응답명", "설명")
# ("OPENAPI00001", 500, "Internal Server Error", "서버 내부 오류")
# ("OPENAPI00002", 403, "Forbidden", "권한이 없는 경우")
# ("OPENAPI00003", 400, "Bad Request", "유효하지 않은 식별자")
# ("OPENAPI00004", 400, "Bad Request", "파라미터 누락 또는 유효하지 않음")
# ("OPENAPI00005", 400, "Bad Request", "유효하지 않은 API KEY")
# ("OPENAPI00006", 400, "Bad Request", "유효하지 않은 게임 또는 API PATH")
# ("OPENAPI00007", 429, "Too Many Requests", "API 호출량 초과")
# ("OPENAPI00009", 400, "Bad Request", "데이터 준비 중")


# code = "OPENAPI00001"

# # code 변수의 값으로 APIErrorCode 열거형 멤버 생성
# # error_code = APIErrorCode(code)

# # # 생성된 열거형 멤버 출력
# # print(error_code)

# # print(APIErrorCode._name_)
# print(APIErrorCode._member_names_)
# print(APIErrorCode._value2member_map_)
# print(APIErrorCode._member_map_)
# print(APIErrorCode.from_code(code))



from enum import Enum

import maplestory.utils.kst as kst


class BaseEnum(Enum):
    """확장된 Enum 클래스로 키와 값에 대한 메서드를 제공합니다."""

    @classmethod
    def keys(cls) -> list[str]:
        """Enum 클래스의 모든 키를 리스트로 반환합니다.

        Returns:
            list: Enum 클래스의 모든 키를 포함하는 리스트입니다.
        """

        return list(cls.__members__)

    @classmethod
    def values(cls) -> list:
        """Enum 클래스의 모든 값을 리스트로 반환합니다.

        Returns:
            list: Enum 클래스의 모든 값을 포함하는 리스트입니다.
        """

        return [c.value for c in cls]


class WorldType(BaseEnum):
    """게임 내의 월드 타입을 나타내는 Enum 클래스입니다."""

    일반 = 0
    리부트 = 1


class GuildRankType(BaseEnum):
    """길드 랭킹 타입을 나타내는 Enum 클래스입니다."""

    주간명성치 = 0
    플레그레이스 = 1
    지하수로 = 2


class DojangDifficulty(BaseEnum):
    """무릉도장 구간를 나타내는 Enum 클래스입니다."""

    일반 = 0  # 105 ~ 200 레벨 (입문?)
    통달 = 1  # 201 이상 레벨


class QueryableDate(BaseEnum):
    """각 카테고리별 조회 가능한 최소 날짜를 나타내는 Enum 클래스입니다."""

    캐릭터 = kst.datetime(2023, 12, 21)
    유니온 = kst.datetime(2023, 12, 21)
    길드 = kst.datetime(2023, 12, 21)
    랭킹 = kst.datetime(2023, 12, 22)
    큐브 = kst.datetime(2022, 11, 25)
    스타포스 = kst.datetime(2023, 12, 27)
    잠재능력 = kst.datetime(2024, 1, 25)



def korean_format_number(n: int) -> str:
    """정수를 한글 형식으로 포맷팅된 문자열로 변환합니다.

    인수:
        n (int): 포맷팅할 정수입니다.

    반환값:
        str: 정수의 한글 형식으로 포맷팅된 문자열입니다.

    예시:
        >>> korean_format_number(12345)
        '1만 2345'

        >>> korean_format_number(27439263548400)
        '27조 4392억 6354만 8400'

    참고:
        이 함수는 정수를 1만 단위로 나누어 한글 형식으로 포맷팅된 문자열로 변환합니다. 사용되는 단위는 "만", "억", "조", "경"입니다.
    """
    """Converts an integer into a Korean formatted string representation.

    Args:
        n (int): The integer to be formatted.

    Returns:
        str: The Korean formatted string representation of the integer.

    Example:
        >>> korean_format_number(12345)
        '1만 2345'

        >>> korean_format_number(27439263548400)
        '27조 4392억 6354만 8400'

    Note:
        This function converts an integer into a Korean formatted string representation by dividing the number into units of 10,000 and appending the corresponding unit name. The units used are "만" (10,000), "억" (100 million), "조" (1 trillion), and "경" (10 quadrillion).
    """

    units = ["", "만", "억", "조", "경"]
    result = []
    unit_index = 0

    while n > 0:
        # 10000으로 나눈 나머지를 구하여 현재 단위의 숫자를 얻습니다.
        part = n % 10000
        if part > 0:  # 현재 단위의 값이 0이 아닌 경우에만 결과에 추가합니다.
            result.insert(0, f"{part}{units[unit_index]}")
        n //= 10000  # 다음 단위로 넘어갑니다.
        unit_index += 1

    return " ".join(result) if result else "0"



def korean_format_number(n: int) -> str:
    """정수를 한글 형식으로 포맷팅된 문자열로 변환합니다.

    Args:
        n (int): 포맷팅할 정수입니다.

    Returns:
        str: 정수의 한글 형식으로 포맷팅된 문자열입니다.

    Examples:
        >>> korean_format_number(12345)
        '1만 2345'

        >>> korean_format_number(27439263548400)
        '27조 4392억 6354만 8400'

    References:
        이 함수는 정수를 1만 단위로 나누어 한글 형식으로 포맷팅된 문자열로 변환합니다.
        사용되는 단위는 "만", "억", "조", "경"입니다.
    """

    units = ["", "만", "억", "조", "경"]
    result = []
    unit_index = 0

    while n > 0:
        # 10000으로 나눈 나머지를 구하여 현재 단위의 숫자를 얻습니다.
        part = n % 10000
        if part > 0:  # 현재 단위의 값이 0이 아닌 경우에만 결과에 추가합니다.
            result.insert(0, f"{part}{units[unit_index]}")
        n //= 10000  # 다음 단위로 넘어갑니다.
        unit_index += 1

    return " ".join(result) if result else "0"



from .character import *
from .guild import *
from .history import *
from .ranking import *
from .union import *

# __all__ = (
#     "Character",
#     "get_character_id",
#     "get_basic_character_info_by_ocid",
#     "get_basic_character_info",
#     "get_popularity_by_ocid",
#     "get_popularity",
#     "get_character_stat_by_ocid",
#     "get_character_stat",
#     "get_character_hyper_stat_by_ocid",
#     "get_character_hyper_stat",
#     "get_character_propensity_by_ocid",
#     "get_character_propensity",
#     "get_character_ability_by_ocid",
#     "get_character_ability",
#     "get_character_equipment_by_ocid",
#     "get_character_equipment",
#     "get_character_cashitem_equipment_by_ocid",
#     "get_character_cashitem_equipment",
#     "get_character_symbol_equipment_by_ocid",
#     "get_character_symbol_equipment",
#     "get_character_set_effect_by_ocid",
#     "get_character_set_effect",
#     "get_character_beauty_equipment_by_ocid",
#     "get_character_beauty_equipment",
#     "get_character_android_equipment_by_ocid",
#     "get_character_android_equipment",
#     "get_character_pet_equipment_by_ocid",
#     "get_character_pet_equipment",
#     "get_character_skill_by_ocid",
#     "get_character_skill",
#     "get_character_link_skill_by_ocid",
#     "get_character_link_skill",
#     "get_character_vmatrix_by_ocid",
#     "get_character_vmatrix",
#     "get_character_hexamatrix_by_ocid",
#     "get_character_hexamatrix",
#     "get_character_hexamatrix_stat_by_ocid",
#     "get_character_hexamatrix_stat",
#     "get_character_dojang_record_by_ocid",
#     "get_character_dojang_record",
# )



from datetime import datetime

from pydantic import BaseModel, Field

import maplestory.utils.kst as kst

from ..types import CubePotenialGrade, ItemSlotName, PageCursor, PotentialGrade

# KST_Timezone = Timezone("Asia/Seoul")
# PotentialDatetime = Annotated[datetime, KST_Timezone, Field(..., ge=POTENTIAL_MINIMUM_DATETIME)]
POTENTIAL_MINIMUM_DATETIME = kst.datetime(2024, 1, 25)


class PotentialOption(BaseModel):
    """잠재능력 옵션 정보

    Attributes:
        option: 옵션 명
        grade: 옵션 등급

    Examples:
        {
            "value": "최대 MP : +5%",
            "grade": "에픽"
        },
        {
            "value": "올스탯 : +5",
            "grade": "레어"
        },
        {
            "value": "점프력 : +5",
            "grade": "레어"
        }
        {
            "value": "STR : +6",
            "grade": "노멀"
        }
    """

    option: str = Field(alias="value")
    grade: CubePotenialGrade

    @property
    def name(self) -> str:
        return self.option.split(":")[0].strip()

    @property
    def value(self) -> int:
        return int(self.option.split(":")[1].replace("%", "").strip())

    @property
    def percent(self) -> bool:
        return self.option.endswith("%")


class CubeHistoryInfo(BaseModel):
    """큐브 사용 결과 정보

    Attributes:
        id: 큐브 히스토리 식별자
        character_name: 캐릭터명
        date_create: 사용 일시 (KST)
        cube_type: 사용 큐브
            "수상한 에디셔널 큐브", "수상한 큐브", ...
        item_upgrade_result: 사용 결과
        miracle_time_flag: 미라클 타임 적용 여부
            "이벤트 적용되지 않음", ...
        item_equipment_part: 장비 분류
        item_level: 장비 레벨
        target_item: 큐브 사용한 장비
        potential_option_grade: 잠재능력 등급
        additional_potential_option_grade: 에디셔널 잠재능력 등급
        upgrade_guarantee: 천장에 도달하여 확정 등급 상승한 여부
        upgrade_guarantee_count: 현재까지 쌓은 스택
        before_potential_option: 사용 전 잠재능력 옵션
        before_additional_potential_option: 사용 전 에디셔널 잠재능력 옵션
        after_potential_option: 사용 후 잠재능력 옵션
        after_additional_potential_option: 사용 후 에디셔널 잠재능력 옵션
    """

    id: str
    character_name: str
    date_create: datetime
    cube_type: str
    item_upgrade_result: str
    miracle_time_flag: str
    item_equipment_part: ItemSlotName
    item_level: int
    target_item: str
    potential_option_grade: PotentialGrade
    additional_potential_option_grade: PotentialGrade
    upgrade_guarantee: bool
    upgrade_guarantee_count: int
    before_potential_option: list[PotentialOption]
    before_additional_potential_option: list[PotentialOption]
    after_potential_option: list[PotentialOption]
    after_additional_potential_option: list[PotentialOption]

    @property
    def is_miracle_time(self) -> bool:
        return self.miracle_time_flag != "이벤트 적용되지 않음"


class CubeHistory(BaseModel):
    """큐브 사용 결과

    Attributes:
        count: 결과 건 수
        next_cursor: 페이징 처리를 위한 cursor
        history: 큐브 히스토리
    """

    count: int
    next_cursor: PageCursor | None
    history: list[CubeHistoryInfo] = Field(alias="cube_history")


class PotentialHistoryInfo(BaseModel):
    """

    Attributes:
        id: 잠재능력 재설정 히스토리 식별자
        character_name: 캐릭터 명
        date_create: 사용 일시 (KST), '2024-01-25T17:28:31.000+09:00'
        potential_type: 대상 잠재능력 타입 (잠재능력, 에디셔널 잠재능력)
        item_upgrade_result: 사용 결과
        miracle_time_flag: 미라클 타임 적용 여부
        item_equipment_part: 장비 분류
        item_level: 장비 레벨
        target_item: 잠재능력 재설정 장비 명
        potential_option_grade: 잠재능력 등급
        additional_potential_option_grade: 에디셔널 잠재능력 등급
        upgrade_guarantee: 천장에 도달하여 확정 등급 상승한 여부
        upgrade_guarantee_count: 현재까지 쌓은 스택
        before_potential_option: 사용 전 잠재능력 옵션
        before_additional_potential_option: 사용 전 에디셔널 잠재능력 옵션
        after_potential_option: 사용 후 잠재능력 옵션
        after_additional_potential_option: 사용 후 에디셔널 잠재능력 옵션
    """

    id: str
    character_name: str
    date_create: datetime
    potential_type: str
    item_upgrade_result: str
    miracle_time_flag: str
    item_equipment_part: str
    item_level: float
    target_item: str
    potential_option_grade: str
    additional_potential_option_grade: str
    upgrade_guarantee: bool
    upgrade_guarantee_count: float
    before_potential_option: list[PotentialOption]
    before_additional_potential_option: list[PotentialOption]
    after_potential_option: list[PotentialOption]
    after_additional_potential_option: list[PotentialOption]


class PotentialHistory(BaseModel):
    """

    Attributes:
        count: 결과 건 수
        next_cursor: 페이징 처리를 위한 cursor
        history: 잠재능력 재설정 히스토리
    """

    count: float
    next_cursor: PageCursor | None
    history: list[PotentialHistoryInfo] = Field(alias="potential_history")



from datetime import datetime
from enum import Enum

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.enums import GuildRankType, QueryableDate, WorldType
from maplestory.models.ranking import (
    AchievementRanking,
    DojangRanking,
    GuildRanking,
    OverallRanking,
    TheSeedRanking,
    UnionRanking,
)
from maplestory.models.types import JobClass, WorldName
from maplestory.utils.network import fetch

# def validate_ranking_date(date: kst.KSTdatetime | None) -> None:
#     if date is not None and date < RANK_MINIMUM_DATETIME:
#         raise ValueError("date must be after 2023-12-22 KST.")


def get_overall_ranking(
    world: WorldName | None = None,
    world_type: int | WorldType | None = None,
    job_class: JobClass | None = None,
    character_id: str | None = None,
    page_number: int = 1,
    base_date: kst.KSTdatetime = kst.yesterday(),
) -> OverallRanking:
    """
    종합 랭킹 정보를 조회합니다.
    Fetches the overall ranking information.

    Args:
        base_date : 조회 기준일(KST) Example : 2023-12-22
                    Query base date (KST)
        world (str, optional): 월드 명
                               World name
                               Available values : 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3
        world_type (int, optional): 월드 타입 (0:일반, 1:리부트) (기본 값은 0이며, world_name 입력 시 미 반영)
                                    World type (0: normal, 1: reboot) (default is 0 and not reflected when world_name is entered)
        job_class (str, optional): 직업 및 전직
                                   Job and class
                                    Available values : 초보자-전체 전직, 전사-전체 전직, 전사-검사, 전사-파이터, 전사-페이지, 전사-스피어맨, 전사-크루세이더, 전사-나이트, 전사-버서커, 전사-히어로, 전사-팔라딘, 전사-다크나이트, 마법사-전체 전직, 마법사-매지션, 마법사-위자드(불,독), 마법사-위자드(썬,콜), 마법사-클레릭, 마법사-메이지(불,독), 마법사-메이지(썬,콜), 마법사-프리스트, 마법사-아크메이지(불,독), 마법사-아크메이지(썬,콜), 마법사-비숍, 궁수-전체 전직, 궁수-아처, 궁수-헌터, 궁수-사수, 궁수-레인저, 궁수-저격수, 궁수-보우마스터, 궁수-신궁, 궁수-아처(패스파인더), 궁수-에인션트아처, 궁수-체이서, 궁수-패스파인더, 도적-전체 전직, 도적-로그, 도적-어쌔신, 도적-시프, 도적-허밋, 도적-시프마스터, 도적-나이트로드, 도적-섀도어, 도적-세미듀어러, 도적-듀어러, 도적-듀얼마스터, 도적-슬래셔, 도적-듀얼블레이더, 해적-전체 전직, 해적-해적, 해적-인파이터, 해적-건슬링거, 해적-캐논슈터, 해적-버커니어, 해적-발키리, 해적-캐논블래스터, 해적-바이퍼, 해적-캡틴, 해적-캐논마스터, 기사단-전체 전직, 기사단-노블레스, 기사단-소울마스터, 기사단-플레임위자드, 기사단-윈드브레이커, 기사단-나이트워커, 기사단-스트라이커, 기사단-미하일, 아란-전체 전직, 에반-전체 전직, 레지스탕스-전체 전직, 레지스탕스-시티즌, 레지스탕스-배틀메이지, 레지스탕스-와일드헌터, 레지스탕스-메카닉, 레지스탕스-데몬슬레이어, 레지스탕스-데몬어벤져, 레지스탕스-제논, 레지스탕스-블래스터, 메르세데스-전체 전직, 팬텀-전체 전직, 루미너스-전체 전직, 카이저-전체 전직, 엔젤릭버스터-전체 전직, 초월자-전체 전직, 초월자-제로, 은월-전체 전직, 프렌즈 월드-전체 전직, 프렌즈 월드-키네시스, 카데나-전체 전직, 일리움-전체 전직, 아크-전체 전직, 호영-전체 전직, 아델-전체 전직, 카인-전체 전직, 라라-전체 전직, 칼리-전체 전직
        character_id (str, optional): 캐릭터 식별자
                                        Character identifier
        page_number : 페이지 번호
                      Page number

    Note:
        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
          Data can be queried from December 22, 2023.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
          Ranking information for the day can be queried from 8:30 AM.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다.
          Please note that the ocid may change due to game content changes.
          ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
          Please be careful when updating services based on ocid.
    """
    dates.is_valid(date, QueryableDate.랭킹)

    if isinstance(world_type, int):
        try:
            world_type = WorldType(world_type)
        except ValueError as err:
            raise ValueError("world_type must be 0 or 1.") from err

    if isinstance(world_type, WorldType):
        world_type = world_type.value

    path = "/maplestory/v1/ranking/overall"
    query = {
        "date": dates.to_string(base_date),
        "world_name": world,
        "world_type": world_type,
        "class": job_class,
        "ocid": character_id,
        "page": page_number,
    }
    response = fetch(path, query)

    return OverallRanking.model_validate(response)


def get_union_ranking(
    world: WorldName | None = None,
    character_id: str | None = None,
    page_number: int = 1,
    date: kst.KSTdatetime = kst.yesterday(),
) -> UnionRanking:
    """유니온 랭킹 정보를 조회합니다.
    Fetches the union ranking information.

    Args:
        date (datetime): 조회 기준일(KST)
               The reference date for the query (KST).
        world (str, optional): 월드 명
                               The name of the world.
            가능한 값은 (스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3)
            Possible values are: 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3
        character_id (str, optional): 캐릭터 식별자
                                      The identifier of the character.
        page_number (int): 페이지 번호
                           The page number.

    Returns:
        UnionRanking: 유니온 랭킹 정보
                      The union ranking information.

    Note:
        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
          Data can be queried from December 22, 2023.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
          Ranking information for the day can be queried from 8:30 AM.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다.
          Please note that the ocid may change due to game content changes.
          ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
          Please be careful when updating services based on ocid.
    """

    dates.is_valid(date, QueryableDate.랭킹)

    path = "/maplestory/v1/ranking/union"
    query = {
        "date": dates.to_string(dates),
        "world_name": world,
        "ocid": character_id,
        "page": page_number,
    }
    response = fetch(path, query)

    return UnionRanking.model_validate(response)


def get_guild_ranking(
    ranking_type: int | GuildRankType = GuildRankType.주간명성치,
    world: WorldName | None = None,
    guild: str | None = None,
    page_number: int = 1,
    date: kst.KSTdatetime = kst.yesterday(),
) -> GuildRanking:
    """길드 랭킹 정보를 조회합니다.
    Fetches the guild ranking information.

    Args:
        date : 조회 기준일(KST)
               The reference date for the query (KST).
        ranking_type (int | GuildRankType): 랭킹 타입
                                            The type of ranking
            (0:주간 명성치, 1:플래그 레이스, 2:지하 수로)
            (0: Weekly Fame, 1: Flag Race, 2: Underground Waterway).
        world (str, optional): 월드 명
                               The name of the world.
            Possible values are: 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3
        guild (str, optional): 길드 명
                               The name of the guild.
        page_number : 페이지 번호
                      The page number.

    Returns:
        GuildRanking: 길드 랭킹 정보
                      The guild ranking information.

    Note:
        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
          Data can be queried from December 22, 2023.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
          Ranking information for the day can be queried from 8:30 AM.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다.
          Please note that the ocid may change due to game content changes.
          ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
          Please be careful when updating services based on ocid.
    """

    dates.is_valid(date, QueryableDate.랭킹)

    if isinstance(ranking_type, GuildRankType):
        ranking_type = ranking_type.value

    if ranking_type not in [0, 1, 2]:
        raise ValueError("ranking_type must be 0, 1, or 2.")

    # if isinstance(ranking_type, int):
    #     try:
    #         ranking_type = GuildRankType(ranking_type)
    #     except ValueError as err:
    #         raise ValueError("ranking_type must be 0 or 1.") from err

    path = "/maplestory/v1/ranking/guild"
    query = {
        "date": dates.to_string(date),
        "ranking_type": ranking_type,
        "world_name": world,
        "guild_name": guild,
        "page": page_number,
    }
    response = fetch(path, query)

    return GuildRanking.model_validate(response)


def get_dojang_ranking(
    world: str | None = None,
    job_class: str | None = None,
    character_id: str | None = None,
    page_number: int = 1,
    difficulty_level: int = 1,
    date: kst.AwareDatetime = kst.yesterday(),
) -> DojangRanking:
    """무릉도장 랭킹 정보를 조회합니다.
    Fetches the Dojang ranking information.

    Args:
        world (Optional[str]): 월드 명
            The name of the world.
            - 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움, 이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3
        job_class (Optional[str]): 직업 및 전직
                    The name of the job class.
                    - 초보자-전체 전직, 전사-전체 전직, 전사-검사, 전사-파이터, 전사-페이지, 전사-스피어맨, 전사-크루세이더, 전사-나이트, 전사-버서커, 전사-히어로, 전사-팔라딘, 전사-다크나이트, 마법사-전체 전직, 마법사-매지션, 마법사-위자드(불,독), 마법사-위자드(썬,콜), 마법사-클레릭, 마법사-메이지(불,독), 마법사-메이지(썬,콜), 마법사-프리스트, 마법사-아크메이지(불,독), 마법사-아크메이지(썬,콜), 마법사-비숍, 궁수-전체 전직, 궁수-아처, 궁수-헌터, 궁수-사수, 궁수-레인저, 궁수-저격수, 궁수-보우마스터, 궁수-신궁, 궁수-아처(패스파인더), 궁수-에인션트아처, 궁수-체이서, 궁수-패스파인더, 도적-전체 전직, 도적-로그, 도적-어쌔신, 도적-시프, 도적-허밋, 도적-시프마스터, 도적-나이트로드, 도적-섀도어, 도적-세미듀어러, 도적-듀어러, 도적-듀얼마스터, 도적-슬래셔, 도적-듀얼블레이더, 해적-전체 전직, 해적-해적, 해적-인파이터, 해적-건슬링거, 해적-캐논슈터, 해적-버커니어, 해적-발키리, 해적-캐논블래스터, 해적-바이퍼, 해적-캡틴, 해적-캐논마스터, 기사단-전체 전직, 기사단-노블레스, 기사단-소울마스터, 기사단-플레임위자드, 기사단-윈드브레이커, 기사단-나이트워커, 기사단-스트라이커, 기사단-미하일, 아란-전체 전직, 에반-전체 전직, 레지스탕스-전체 전직, 레지스탕스-시티즌, 레지스탕스-배틀메이지, 레지스탕스-와일드헌터, 레지스탕스-메카닉, 레지스탕스-데몬슬레이어, 레지스탕스-데몬어벤져, 레지스탕스-제논, 레지스탕스-블래스터, 메르세데스-전체 전직, 팬텀-전체 전직, 루미너스-전체 전직, 카이저-전체 전직, 엔젤릭버스터-전체 전직, 초월자-전체 전직, 초월자-제로, 은월-전체 전직, 프렌즈 월드-전체 전직, 프렌즈 월드-키네시스, 카데나-전체 전직, 일리움-전체 전직, 아크-전체 전직, 호영-전체 전직, 아델-전체 전직, 카인-전체 전직, 라라-전체 전직, 칼리-전체 전직
        character_id (Optional[str]): 캐릭터 식별자
                    The identifier of the character.
        page_number : 페이지 번호
                    The page number.
        difficulty_level : 구간 (0:일반, 1:통달)
            The difficulty level (0: Normal, 1: Mastery).
        date : 조회 기준일(KST)
            The reference date for the query (KST).

    Returns:
        DojangRanking: The Dojang ranking information.

    Note:
        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
          Data can be queried from December 22, 2023.
        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
          Ranking information for the day can be queried from 8:30 AM.
        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다.
          Please note that the ocid may change due to game content changes.
          ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
          Please be careful when updating services based on ocid.
    """
    dates.is_valid(date, QueryableDate.랭킹)

    path = "/maplestory/v1/ranking/dojang"
    query = {
        "date": dates.to_string(date),
        "world_name": world,
        "class": job_class,
        "ocid": character_id,
        "page": page_number,
        "difficulty": difficulty_level,
    }
    response = fetch(path, query)

    return DojangRanking.model_validate(response)


def get_theseed_ranking(
    world: WorldName | None = None,
    character_id: str | None = None,
    page_number: int = 1,
    date: kst.AwareDatetime = kst.yesterday(),
) -> TheSeedRanking:
    """
    Fetches the TheSeed ranking information.
    더 시드 랭킹 정보를 조회합니다.

    Args:
        world (str): The name of the world.
                     월드명.
        character_id (str): The identifier of the character.
                            캐릭터 식별자(ocid).
        page_number (int): The page number.
                           페이지 번호.
        date (datetime): The reference date for the query (KST).
                                   조회 기준일(KST)

    Returns:
        TheSeedRanking: The TheSeed ranking information.
                        더 시드 랭킹 정보.

    Note:
        - Data can be queried from December 22, 2023.
          2023년 12월 22일 데이터부터 조회할 수 있습니다.
        - Ranking information can be queried from 8:30 AM.
          오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
        - Be careful when updating the service based on ocid, as the ocid may change due to game content changes.
          게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
    """

    dates.is_valid(date, QueryableDate.랭킹)

    path = "/maplestory/v1/ranking/theseed"
    query = {
        "date": dates.to_string(date),
        "world_name": world,
        "ocid": character_id,
        "page": page_number,
    }
    response = fetch(path, query)

    return TheSeedRanking.model_validate(response)


def get_achievement_ranking(
    character_id: str | None = None,
    page_number: int = 1,
    date: kst.AwareDatetime = kst.yesterday(),
) -> AchievementRanking:
    """
    Fetches the achievement ranking information.
    조회 기준일(KST)과 캐릭터 식별자를 사용하여 업적 랭킹 정보를 조회합니다.

    Args:
        character_id (str): 캐릭터 식별자(ocid).
                            The identifier of the character.

        page_number (int): 페이지 번호.
                           The page number.

        date (datetime): 조회 기준일(KST)
                                   The reference date for the query (KST).

    Returns:
        AchievementRanking: 업적 랭킹 정보.
                            The achievement ranking information.

    Note:
        - 2023년 12월 22일 데이터부터 조회할 수 있습니다.
          Data can be queried from December 22, 2023.

        - 오전 8시 30분부터 오늘의 랭킹 정보를 조회할 수 있습니다.
          Ranking information can be queried from 8:30 AM.

        - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
          Be careful when updating the service based on ocid, as the ocid may change due to game content changes.

    """

    dates.is_valid(date, QueryableDate.랭킹)

    path = "/maplestory/v1/ranking/achievement"
    query = {
        "date": dates.to_string(date),
        "ocid": character_id,
        "page": page_number,
    }
    response = fetch(path, query)

    return AchievementRanking.model_validate(response)



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field, HttpUrl, field_validator

from ..types import AndroidGender, ExpireDate, ItemGender, LabelName, PresetNumber


class AndroidHair(BaseModel):
    """안드로이드 헤어 정보

    Attributes:
        hair_name: 안드로이드 헤어 명
        base_color: 안드로이드 헤어 베이스 컬러
        mix_color: 안드로이드 헤어 믹스 컬러
        mix_rate: 안드로이드 헤어 믹스 컬러의 염색 비율

    Examples:
        "hair_name": "검은색 방울꽃 헤어",
        "base_color": "검은색",
        "mix_color": null,
        "mix_rate": "0"
    """

    hair_name: str | None
    base_color: str | None
    mix_color: str | None
    mix_rate: int


class AndroidFace(BaseModel):
    """안드로이드 성형 정보

    Attributes:
        face_name: 안드로이드 성형 명
        base_color: 안드로이드 성형 베이스 컬러
        mix_color: 안드로이드 성형 믹스 컬러
        mix_rate: 안드로이드 성형 믹스 컬러의 염색 비율

    Examples:
        "face_name": "어제 울었어 얼굴",
        "base_color": "검은색",
        "mix_color": null,
        "mix_rate": "0"
    """

    face_name: str | None
    base_color: str | None
    mix_color: str | None
    mix_rate: int


class AndroidCashItemColoringPrism(BaseModel):
    """안드로이드 캐시아이템 컬러링프리즘 정보

    Attributes:
        color_range: 컬러링프리즘 색상 범위
        hue: 컬러링프리즘 색조
        saturation: 컬러링프리즘 채도
        value: 컬러링프리즘 명도
    """

    color_range: str
    hue: int
    saturation: int
    value: int


class AndroidCashItemOption(BaseModel):
    """안드로이드 캐시아이템 옵션

    Attributes:
        type: 옵션 타입
        value: 옵션 값
    """

    type: str = Field(alias="option_type")
    value: str = Field(alias="option_value")


class AndroidCashItem(BaseModel):
    """안드로이드 캐시 아이템 장착 정보

    Attributes:
        part: 안드로이드 캐시 아이템 부위 명
        slot: 안드로이드 캐시 아이템 슬롯 위치
        name: 안드로이드 캐시 아이템 명
        icon: 안드로이드 캐시 아이템 아이콘
        description: 안드로이드 캐시 아이템 옵션
        option: 안드로이드 캐시 아이템 옵션
        date_expire: 안드로이드 캐시 아이템 유효기간 (KST)
        date_option_expire: 안드로이드 캐시 아이템 옵션 유효기간 (KST, 시간 단위 데이터로 분은 일괄 0으로 표기)
        label: 안드로이드 캐시 아이템 라벨 정보 (스페셜라벨, 레드라벨, 블랙라벨, 마스터라벨)
        coloring_prism: 안드로이드 아이템 컬러링프리즘 정보
        gender: 아이템 장착 가능 성별
    """

    part: str = Field(alias="cash_item_equipment_part")
    slot: str = Field(alias="cash_item_equipment_slot")
    name: str = Field(alias="cash_item_name")
    icon: str = Field(alias="cash_item_icon")
    description: str | None = Field(alias="cash_item_description")
    option: list[AndroidCashItemOption] = Field(alias="cash_item_option")
    date_expire: ExpireDate
    date_option_expire: ExpireDate
    label: LabelName = Field(alias="cash_item_label")
    coloring_prism: AndroidCashItemColoringPrism | None = Field(
        alias="cash_item_coloring_prism"
    )
    gender: ItemGender = Field(alias="android_item_gender")


class AndroidPreset(BaseModel):
    """안드로이드 프리셋 정보

    Attributes:
        name: 안드로이드 명
        nickname: 안드로이드 닉네임
        icon: 안드로이드 아이콘
        description: 안드로이드 아이템 설명
        gender: 안드로이드 성별
        grade: 안드로이드 등급
        skin_name: 안드로이드 피부
        hair: 안드로이드 헤어 정보
        face: 안드로이드 성형 정보
        ear_sensor_clip_flag: 안드로이드 이어센서 클립 적용 여부
        non_humanoid_flag: 비인간형 안드로이드 여부
        shop_usable_flag: 잡화상점 기능 이용 가능 여부
    """

    name: str = Field(alias="android_name")
    nickname: str | None = Field(alias="android_nickname")
    icon: str = Field(alias="android_icon")
    description: str = Field(alias="android_description")
    gender: str = Field(alias="android_gender")
    grade: str = Field(alias="android_grade")
    skin_name: str | None = Field(alias="android_skin_name")
    hair: AndroidHair = Field(alias="android_hair")
    face: AndroidFace = Field(alias="android_face")
    ear_sensor_clip_flag: str = Field(alias="android_ear_sensor_clip_flag")
    non_humanoid_flag: str = Field(alias="android_non_humanoid_flag")
    shop_usable_flag: str = Field(alias="android_shop_usable_flag")


# FIXME: android_non_humanoid_flag가 null로 잘못 나오는 경우가 있음.
# FIXME: android_shop_usable_flag가 null로 잘못 나오는 경우가 있음.
# FIXME: android_preset_1에는 인간형, 가능인데, 현재 적용되어 있는 값에는 null, null임.
class AndroidEquipment(BaseModel):
    """안드로이드 장비 정보

    Attributes:
        date: 조회 기준일 (KST)
        name: 안드로이드 명
        nickname: 안드로이드 닉네임
        icon: 안드로이드 아이콘
        description: 안드로이드 아이템 설명
        hair: 안드로이드 헤어 정보
        face: 안드로이드 성형 정보
        skin_name: 안드로이드 피부 명
        cash_item_equipment: 안드로이드 캐시 아이템 장착 정보
        ear_sensor_clip_flag: 안드로이드 이어센서 클립 적용 여부 (미적용, 적용)
        gender: 안드로이드 성별
        grade: 안드로이드 등급
        non_humanoid_flag: 비인간형 안드로이드 여부 (인간형, 비인간형)
        shop_usable_flag: 잡화상점 기능 이용 가능 여부 (가능, 불가능)
        preset_no: 적용 중인 장비 프리셋 번호(number)
        preset_1: 1번 프리셋 안드로이드 정보
        preset_2: 2번 프리셋 안드로이드 정보
        preset_3: 3번 프리셋 안드로이드 정보
    """

    date: datetime
    name: str | None = Field(alias="android_name")
    nickname: str | None = Field(alias="android_nickname")
    icon: HttpUrl | None = Field(alias="android_icon")
    description: str | None = Field(alias="android_description")
    hair: AndroidHair = Field(alias="android_hair")
    face: AndroidFace = Field(alias="android_face")
    skin_name: str | None = Field(alias="android_skin_name")
    cash_item_equipment: list[AndroidCashItem] = Field(
        alias="android_cash_item_equipment"
    )
    ear_sensor_clip: bool = Field(alias="android_ear_sensor_clip_flag")
    gender: AndroidGender = Field(alias="android_gender")
    grade: str | None = Field(alias="android_grade")
    non_humanoid: str | None = Field(alias="android_non_humanoid_flag")
    shop_usable: bool | None = Field(alias="android_shop_usable_flag")
    preset_no: PresetNumber
    preset_1: AndroidPreset | None = Field(alias="android_preset_1")
    preset_2: AndroidPreset | None = Field(alias="android_preset_2")
    preset_3: AndroidPreset | None = Field(alias="android_preset_3")

    @field_validator("ear_sensor_clip", mode="before")
    @classmethod
    def change_ear_sensor_clip(cls, v: Any) -> bool | None:
        match v:
            case "미적용" | "0" | 0:
                return False
            case "적용" | "1" | 1:
                return True
            case _:
                return v

    @field_validator("shop_usable", mode="before")
    @classmethod
    def change_shop_usable(cls, v: Any) -> bool | None:
        match v:
            case "불가능" | "0" | 0:
                return False
            case "가능" | "1" | 1:
                return True
            case _:
                return v

    @property
    def is_humanoid(self) -> bool:
        return self.non_humanoid == "인간형"


# E       pydantic_core._pydantic_core.ValidationError: 2 validation errors for AndroidEquipment
# E       android_ear_sensor_clip_flag
# E         Input should be a valid string [type=string_type, input_value=False, input_type=bool]
# E           For further information visit https://errors.pydantic.dev/2.5/v/string_type
# E       android_shop_usable_flag
# E         Value error, Unexpected value on shop_usable_flag: None [type=value_error, input_value=None, input_type=NoneType]
# E           For further information visit https://errors.pydantic.dev/2.5/v/value_error



from datetime import datetime

from pydantic import BaseModel, Field, HttpUrl

from ..types import CharacterClass, LinkSkillLevel


class LinkSkill(BaseModel):
    """링크 스킬 정보

    Attributes:
        name (str): 스킬 명
        description (str): 스킬 설명
        level (LinkSkillLevel): 스킬 레벨
        effect (str): 스킬 효과
        icon (HttpUrl): 스킬 아이콘
    """

    name: str = Field(alias="skill_name")
    description: str = Field(alias="skill_description")
    level: LinkSkillLevel = Field(alias="skill_level")
    effect: str = Field(alias="skill_effect")
    icon: HttpUrl = Field(alias="skill_icon")


class CharacterLinkSkill(BaseModel):
    """캐릭터 링크 스킬 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class (CharacterClass): 캐릭터 직업
        link_skill (list[LinkSkill]): 링크 스킬 정보
        owned_link_skill (LinkSkill | None): 내 링크 스킬 정보
        link_skill_preset_1 (list[LinkSkill]): 링크 스킬 1번 프리셋 정보
        link_skill_preset_2 (list[LinkSkill]): 링크 스킬 2번 프리셋 정보
        link_skill_preset_3 (list[LinkSkill]): 링크 스킬 3번 프리셋 정보
        owned_link_skill_preset_1 (LinkSkill | None): 내 링크 스킬 1번 프리셋 정보
        owned_link_skill_preset_2 (LinkSkill | None): 내 링크 스킬 2번 프리셋 정보
        owned_link_skill_preset_3 (LinkSkill | None): 내 링크 스킬 3번 프리셋 정보
    """

    date: datetime = Field(repr=False)
    character_class: CharacterClass
    link_skill: list[LinkSkill] = Field(alias="character_link_skill")
    owned_link_skill: LinkSkill = Field(alias="character_owned_link_skill")
    link_skill_preset_1: list[LinkSkill] = Field(alias="character_link_skill_preset_1")
    link_skill_preset_2: list[LinkSkill] = Field(alias="character_link_skill_preset_2")
    link_skill_preset_3: list[LinkSkill] = Field(alias="character_link_skill_preset_3")
    owned_link_skill_preset_1: LinkSkill | None = Field(
        alias="character_owned_link_skill_preset_1"
    )
    owned_link_skill_preset_2: LinkSkill | None = Field(
        alias="character_owned_link_skill_preset_2"
    )
    owned_link_skill_preset_3: LinkSkill | None = Field(
        alias="character_owned_link_skill_preset_3"
    )



from datetime import datetime
from typing import Annotated, Literal

from annotated_types import Ge, Le, Len

ExpireDate = datetime | None
Grade = Literal["레전드리", "유니크", "에픽", "레어"]
PageCursor = Annotated[str, Len(64, 64)]
PresetNumber = Annotated[int, Ge(1), Le(3)] | None
WorldName = Literal[
    "스카니아",
    "베라",
    "루나",
    "제니스",
    "크로아",
    "유니온",
    "엘리시움",
    "이노시스",
    "레드",
    "오로라",
    "아케인",
    "노바",
    "리부트",
    "리부트2",
    "버닝",
    "버닝2",
    "버닝3",
]
